# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import atexit
import os
import threading
import traceback

from msprobe.core.common.const import Const
from msprobe.core.common.log import logger
from msprobe.core.common.megatron_utils import MegatronStepInfo, get_micro_step, is_megatron
from msprobe.core.dump.data_dump.data_processor.factory import DataProcessorFactory
from msprobe.core.dump.data_dump.json_writer import DataWriter
from msprobe.core.dump.data_dump.scope import ScopeFactory


def build_data_collector(config):
    return DataCollector(config)


class DataCollector:
    tasks_need_tensor_data = [Const.DIFF_CHECK, Const.TENSOR]
    level_without_construct = [Const.LEVEL_L1, Const.LEVEL_L2]

    def __init__(self, config):
        self.config = config
        self.data_writer = DataWriter()
        self.data_processor = DataProcessorFactory.create_processor(self.config, self.data_writer)
        self.module_processor = DataProcessorFactory.get_module_processor(self.config.framework)
        self.module_count = {}
        self.scope = ScopeFactory(self.config).build_scope()
        self.backward_module_names = {}
        self.params_grad_record = {}
        self.optimizer_status = ""
        self.optimizer_status_first_start = {Const.OPTIMIZER: True, Const.CLIP_GRAD: True}
        atexit.register(self.write_json_at_exit)

    @property
    def dump_data_dir(self):
        return self.data_writer.dump_tensor_data_dir

    @property
    def dump_file_path(self):
        return self.data_writer.dump_file_path

    @property
    def bench_dump_file_path(self):
        return self.data_writer.bench_dump_file_path

    @staticmethod
    def check_scope_and_pid(scope, name, pid):
        return (not scope or scope.check(name)) and pid == os.getpid()

    @staticmethod
    def set_is_recomputable(data_info, is_recompute):
        if data_info and len(data_info) == 1 and is_recompute is not None: # 正常情况下data_info的长度应改为1
            data_info[list(data_info.keys())[0]]["is_recompute"] = is_recompute

    def reset_status(self):
        self.optimizer_status = ""
        self.optimizer_status_first_start = {Const.OPTIMIZER: True, Const.CLIP_GRAD: True}
        self.data_writer.reset_cache()
        self.backward_module_names.clear()

    def update_api_or_module_name(self, api_or_module_name):
        self.data_processor.update_api_or_module_name(api_or_module_name)

    def write_json(self):
        self.data_writer.write_json()

    def write_json_at_exit(self):
        if self.config.async_dump and self.config.task == Const.TENSOR:
            self.data_processor.dump_async_data()
        self.data_writer.write_json()

    def update_data(self, name, data_info):
        logger.debug(f"msprobe is collecting data on {name}.")
        self.data_writer.update_data(data_info)

    def call_stack_collect(self, data_info, name):
        stack_info, is_recompute = self.data_processor.analyze_api_call_stack(name)
        self.data_writer.update_stack(name, stack_info)
        self.set_is_recomputable(data_info, is_recompute)

    def forward_input_data_collect(self, name, module, pid, module_input_output):
        try:
            if not self.check_scope_and_pid(self.scope, name, pid):
                return

            data_info = {}
            if self.config.task != Const.STRUCTURE:
                data_info = self.data_processor.analyze_forward_input(name, module, module_input_output)
            if self.config.level == Const.LEVEL_L2:
                return
            self.call_stack_collect(data_info, name)
            self.handle_data(name, data_info, flush=self.data_processor.is_terminated)

        except Exception as e:
            # 取异常类名作为“类型”做去重
            error_type = type(e).__name__
            tb = traceback.format_exc()
            self.data_writer.write_error_log(
                f"[ERROR] forward_input_data_collect failed: name={name}, pid={pid}\n{tb}",
                error_type=error_type
            )

    def forward_output_data_collect(self, name, module, pid, module_input_output):
        self.update_construct(name)
        if not self.check_scope_and_pid(self.scope, name, pid):
            return

        data_info = {}
        if self.config.task != Const.STRUCTURE:
            data_info = self.data_processor.analyze_forward_output(name, module, module_input_output)
        is_recompute = self.data_processor.is_recompute()
        self.set_is_recomputable(data_info, is_recompute)
        if self.config.level == Const.LEVEL_L2:
            return
        self.handle_data(name, data_info, flush=self.data_processor.is_terminated)

    def forward_data_collect_only_tensor(self, name, module, pid, module_input_output):
        if not self.check_scope_and_pid(self.scope, name, pid):
            return
        self.data_processor.analyze_forward(name, module, module_input_output)


    def forward_data_collect(self, name, module, pid, module_input_output):
        self.update_construct(name)
        if not self.check_scope_and_pid(self.scope, name, pid):
            return
        data_info = {}
        if self.config.task != Const.STRUCTURE:
            data_info = self.data_processor.analyze_forward(name, module, module_input_output)
        self.call_stack_collect(data_info, name)
        self.handle_data(name, data_info, flush=self.data_processor.is_terminated)


    def backward_data_collect_only_tensor(self, name, module, pid, module_input_output):
        if not self.check_scope_and_pid(self.scope, name, pid):
            return
        self.data_processor.analyze_backward(name, module, module_input_output)


    def backward_data_collect(self, name, module, pid, module_input_output):
        self.update_construct(name)
        if not self.check_scope_and_pid(self.scope, name, pid):
            return
        data_info = {}
        if self.config.task != Const.STRUCTURE:
            data_info = self.data_processor.analyze_backward(name, module, module_input_output)
        if self.config.level == Const.LEVEL_L2:
            return
        if data_info and name.split(Const.SEP)[0] in Const.MODULE_PREFIX:
            module_name = name.rsplit(Const.SEP, 2)[0]
            self.backward_module_names[module_name] = True
        self.handle_data(name, data_info, flush=self.data_processor.is_terminated)


    def backward_input_data_collect(self, name, module, pid, module_input_output):
        self.update_construct(name)
        if not self.check_scope_and_pid(self.scope, name, pid):
            return
        data_info = {}
        if self.config.task != Const.STRUCTURE:
            data_info = self.data_processor.analyze_backward_input(name, module, module_input_output)
        self.handle_data(name, data_info)


    def backward_output_data_collect(self, name, module, pid, module_input_output):
        self.update_construct(name)
        if not self.check_scope_and_pid(self.scope, name, pid):
            return
        data_info = {}
        if self.config.task != Const.STRUCTURE:
            data_info = self.data_processor.analyze_backward_output(name, module, module_input_output)
        self.handle_data(name, data_info)


    def update_construct(self, name):
        if self.config.level not in DataCollector.level_without_construct:
            if self.optimizer_status in [Const.OPTIMIZER, Const.CLIP_GRAD]:
                if self.optimizer_status_first_start[self.optimizer_status]:
                    self.data_writer.update_construct(
                        {self.optimizer_status: None if not is_megatron() else [None, get_micro_step()]})
                    self.optimizer_status_first_start[self.optimizer_status] = False
                self.data_writer.update_construct(
                    {name: self.optimizer_status if not is_megatron() else [self.optimizer_status, get_micro_step()]})
            else:
                if self.config.level == Const.LEVEL_MIX and \
                  not (name.startswith(Const.MODULE) or name.startswith(Const.CELL)):
                    self.data_writer.update_construct(
                        {name: self.module_processor.api_parent_node.get(threading.get_ident())}
                    )
            if MegatronStepInfo.is_megatron:
                micro_step_number = max(MegatronStepInfo.forward_micro_step, MegatronStepInfo.backward_micro_step)
                self.data_writer.update_construct({Const.MEGATRON_MICRO_STEP_NUMBER: micro_step_number})

            self.data_writer.update_construct(self.module_processor.module_node)

    def handle_data(self, name, data_info, flush=False):
        if data_info:
            self.update_data(name, data_info)
        if self.config.async_dump:
            return
        if not flush:
            self.data_writer.flush_data_periodically()
        else:
            self.write_json()

    def update_dump_paths(self, *args):
        self.data_writer.update_dump_paths(*args)

    def initialize_json_file(self, framework=Const.UNKNOWN_FRAMEWORK):
        self.data_writer.initialize_json_file(task=self.config.task, level=self.config.level, framework=framework)

    def update_iter(self, current_iter):
        self.data_processor.update_iter(current_iter)

    def params_data_collect(self, name, param_name, pid, data, index):
        grad_name = name + Const.SEP + Const.PARAMS_GRAD + Const.SEP + index
        self.update_api_or_module_name(grad_name)
        if not self.check_scope_and_pid(self.scope, name, pid) and not self.backward_module_names.get(name):
            if self.data_writer.cache_data.get("data"):
                self.data_writer.cache_data.get("data").pop(grad_name, None)
                self.params_grad_record[grad_name] = False
            return
        data_info = self.data_processor.analyze_params(grad_name, param_name, data)
        self.handle_data(grad_name, data_info, flush=self.data_processor.is_terminated)
        self.params_grad_record[grad_name] = False

    def debug_data_collect_forward(self, variable, name_with_count):
        data_info = self.data_processor.analyze_debug_forward(variable, name_with_count)
        name_with_count_category = name_with_count + Const.SEP + Const.DEBUG
        self.data_writer.update_debug({name_with_count_category: data_info})

    def debug_data_collect_backward(self, variable, grad_name_with_count):
        # prepare all None nested data structure
        all_none_data_info = self.data_processor.analyze_element_to_all_none(variable)
        grad_name_with_count_category = grad_name_with_count + Const.SEP + Const.DEBUG
        self.data_writer.update_debug({grad_name_with_count_category: all_none_data_info})

        # register tensor backward hook
        self.data_processor.analyze_debug_backward(variable, grad_name_with_count_category,
                                                   self.data_writer.cache_debug['data'])



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
from collections import defaultdict, namedtuple

from mindspore.ops.operations import _inner_ops as inner
from mindspore._c_expression import MSContext

from msprobe.core.common.const import Const, MsgConst
from msprobe.core.common.utils import check_token_range, ThreadSafe, check_rank_id
from msprobe.core.common.runtime import Runtime
from msprobe.core.dump.debugger.precision_debugger import BasePrecisionDebugger
from msprobe.mindspore.dump.cell_processor import CellProcessor
from msprobe.mindspore.common.const import Const as MsConst
from msprobe.mindspore.common.utils import (
    set_register_backward_hook_functions,
    check_save_param,
    is_graph_mode_cell_dump_allowed,
    wrap_backward_hook_call_func
)
from msprobe.mindspore.dump.debugger.debugger_config import DebuggerConfig
from msprobe.mindspore.dump.dump_processor.graph_mode_cell_dump import GraphModeCellDump
from msprobe.mindspore.dump.dump_processor.hook_cell.api_register import get_api_register
from msprobe.mindspore.dump.dump_processor.hook_cell.hook_cell import HOOKCell
from msprobe.mindspore.dump.ms_config import parse_task_config
from msprobe.mindspore.dump.mindspore_service import MindsporeService
from msprobe.mindspore.dump.task_handler_factory import TaskHandlerFactory

try:
    import mindspore as ms
    from mindspore._c_expression import _dump_start, _dump_stop, _dump_step, _set_init_iter, _dump_set_dynamic
except ImportError:
    enable_dynamic_kbyk_dump = False
else:
    enable_dynamic_kbyk_dump = True

try:
    from msprobe.lib import _msprobe_c
except ImportError:
    _msprobe_c = None


ConfigParameters = namedtuple("ConfigParameters", ["config_path", "task", "dump_path", "level"])


class PrecisionDebugger(BasePrecisionDebugger):

    def __init__(
            self,
            config_path=None,
            task=None,
            dump_path=None,
            level=None,
            step=None
    ):
        if self.initialized:
            return
        set_register_backward_hook_functions()
        super().__init__(config_path, task, dump_path, level, step)

        self.common_config.level = level if level else self.common_config.level
        self.common_config.dump_path = dump_path if dump_path else self.common_config.dump_path
        self.config = DebuggerConfig(self.common_config, self.task_config)

        if self._is_kernel_dump() and not self.task_config.is_regex_valid:
            raise ValueError('Illegal regular expressions exist in the list.')

        setattr(inner.CellBackwardHook, '__call__',
                wrap_backward_hook_call_func(getattr(inner.CellBackwardHook, '__call__')))

        if self._is_kernel_dump() and _msprobe_c:
            os.environ["MS_HOOK_ENABLE"] = "on"
            _msprobe_c._PrecisionDebugger(framework="MindSpore", config_path=config_path)

        self.config.execution_mode = self._get_execution_mode()
        if self._need_service():
            self.service = MindsporeService(self.config)

        Runtime.step_count = 0
        Runtime.is_running = False
        if enable_dynamic_kbyk_dump and self.config.level_ori == Const.LEVEL_L2:
            _dump_set_dynamic()

    @staticmethod
    def _get_task_config(task, json_config):
        return parse_task_config(task, json_config)

    @staticmethod
    def _get_execution_mode():
        jit_level = ms.context.get_jit_config().get(MsConst.JIT_LEVEL)
        jit_level = jit_level if jit_level else ms.get_context(MsConst.JIT_LEVEL)
        if not jit_level:
            if MSContext.get_instance().get_ascend_soc_version() == MsConst.ASCEND_910A:
                jit_level = MsConst.JIT_LEVEL_O2
            else:
                jit_level = MsConst.JIT_LEVEL_O0

        if ms.get_context("mode") == ms.GRAPH_MODE:
            if jit_level == MsConst.JIT_LEVEL_O2:
                return MsConst.GRAPH_GE_MODE
            else:
                return MsConst.GRAPH_KBYK_MODE
        else:
            return MsConst.PYNATIVE_MODE

    @staticmethod
    def _is_graph_dump(config: DebuggerConfig):
        if not config.list:
            return True
        is_graph = any(item.startswith("name-regex") for item in config.list)
        is_graph |= all("." not in item for item in config.list)
        return is_graph

    @classmethod
    def handle_graph_cell_dump(cls, instance):
        """处理graph cell dump相关逻辑"""
        is_graph_mode = instance.config.execution_mode in [MsConst.GRAPH_GE_MODE, MsConst.GRAPH_KBYK_MODE]

        if is_graph_mode:
            GraphModeCellDump.step(instance.config.dump_path, instance.config.step, instance.config.task)

        if Runtime.run_mode == MsConst.PYNATIVE_GRAPH_MODE:
            dump_path = os.path.join(instance.config.dump_path, MsConst.GRAPH_MODE)
            GraphModeCellDump.step(dump_path, instance.config.step, instance.config.task)

    @classmethod
    def start(cls, model=None, token_range=None, rank_id=None):
        instance = cls._get_instance()
        if instance is None:
            return
        check_rank_id(rank_id)
        if cls._is_kernel_dump():
            cls._start_kernel_dump()
        else:
            check_token_range(token_range)
            instance.config.execution_mode = cls._get_execution_mode()
            if cls._need_service():
                with ThreadSafe():
                    if not instance.service:
                        instance.service = MindsporeService(instance.config)
                    instance.config.check_model(model, token_range)
                    instance.service.start(model, token_range, rank_id)
            else:
                if not instance.first_start:
                    get_api_register().restore_all_api()
                    handler = TaskHandlerFactory.create(instance.config, model)
                    handler.handle()
                Runtime.is_running = True
        instance.first_start = True

    @classmethod
    def stop(cls):
        instance = cls._get_instance()
        if instance is None:
            return

        if instance.service:
            with ThreadSafe():
                instance.service.stop()
        else:
            Runtime.is_running = False
        if enable_dynamic_kbyk_dump and instance.config.level_ori == Const.LEVEL_L2:
            ms.runtime.synchronize()
            _dump_stop()
        if cls._is_kernel_dump() and _msprobe_c:
            _msprobe_c._PrecisionDebugger().stop()

    @classmethod
    def step(cls):
        instance = cls._get_instance()
        if instance is None:
            return

        if instance.service:
            with ThreadSafe():
                instance.service.step()

        if instance.config.level in [MsConst.CELL, MsConst.CELL_AND_API]:
            cls.handle_graph_cell_dump(instance)

        if enable_dynamic_kbyk_dump and instance.config.level_ori == Const.LEVEL_L2:
            _dump_step(1)
        if cls._is_kernel_dump() and _msprobe_c:
            _msprobe_c._PrecisionDebugger().step()

        HOOKCell.cell_count = defaultdict(int)
        CellProcessor.reset_cell_stats()
        Runtime.step_count += 1

    @classmethod
    @ThreadSafe.synchronized
    def save(cls, variable, name, save_backward=True):
        instance = cls._instance
        if not instance:
            raise Exception(MsgConst.NOT_CREATED_INSTANCE)
        if instance.task not in [Const.TENSOR, Const.STATISTICS] or instance.config.level_ori != Const.LEVEL_DEBUG:
            return
        try:
            check_save_param(variable, name, save_backward)
        except ValueError:
            return
        if not instance.service:
            instance.service = MindsporeService(instance.config)
        instance.service.save(variable, name, save_backward)

    @classmethod
    def _need_service(cls):
        instance = cls._instance
        if not instance:
            raise Exception(MsgConst.NOT_CREATED_INSTANCE)
        if instance.config.level_ori == Const.LEVEL_L2:
            return not instance._is_graph_dump(instance.config)
        if instance.config.execution_mode != MsConst.PYNATIVE_MODE:
            return False
        return True

    @classmethod
    def _is_kernel_dump(cls):
        instance = cls._instance
        if not instance:
            raise Exception(MsgConst.NOT_CREATED_INSTANCE)
        return instance.config.level_ori == Const.LEVEL_L2

    @classmethod
    def _start_kernel_dump(cls):
        instance = cls._get_instance()
        is_graph_config = cls._is_graph_dump(instance.config)
        instance.config.check_config_with_l2(is_graph_config)
        if not is_graph_config:
            if not instance.service:
                instance.service = MindsporeService(instance.config)
            instance.service.start()
        else:
            if _msprobe_c:
                _msprobe_c._PrecisionDebugger().start()
            if not instance.first_start:
                get_api_register().restore_all_api()
                handlers = TaskHandlerFactory.create(instance.config)
                for handler in handlers:
                    handler.handle()
                if enable_dynamic_kbyk_dump:
                    _set_init_iter(0)
            if enable_dynamic_kbyk_dump:
                is_valid_rank = (not instance.config.rank or Runtime.rank_id in instance.config.rank)
                is_valid_step = (not instance.config.step or Runtime.step_count in instance.config.step)
                if is_valid_rank and is_valid_step:
                    _dump_start()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from abc import ABC, abstractmethod

import mindspore
import numpy as np
import torch
from msprobe.core.common.const import CompareConst
from msprobe.core.common.exceptions import ApiAccuracyCheckerException
from msprobe.mindspore.common.log import logger
from msprobe.mindspore.common.const import MsCompareConst


class CompareResult:
    def __init__(self, compare_value, pass_status, err_msg):
        self.compare_value = compare_value
        self.pass_status = pass_status
        self.err_msg = err_msg


class BaseCompareAlgorithm(ABC):
    def __init__(self) -> None:
        super().__init__()
        self.compare_algorithm_name = None
        self.err_msg_mapping = {
            CompareConst.COSINE: {
                CompareConst.PASS: "",
                CompareConst.ERROR: f"cosine similarity is less than threshold: {CompareConst.COS_THRESHOLD} ",
                CompareConst.SKIP: "two inputs are not valid for computing cosine similarity, skip comparing ",
            },
            CompareConst.MAX_ABS_ERR: {
                CompareConst.PASS: "",
                CompareConst.ERROR: "max absolute difference is greater than " \
                                    f"threshold: {CompareConst.MAX_ABS_ERR_THRESHOLD} ",
                CompareConst.SKIP: "two inputs are not valid for computing max absolute difference, skip comparing ",
            },
            CompareConst.MAX_RELATIVE_ERR: {
                CompareConst.PASS: "",
                CompareConst.ERROR: "",
                CompareConst.SKIP: "",
            },
        }

    def __call__(self, bench_compute_element, tested_compute_element):
        '''
        Args:
            bench_compute_element: ComputeElement
            tested_compute_element: ComputeElement

        Return:
            compare_result: CompareResult
        '''
        if self.check_validity(bench_compute_element, tested_compute_element):
            compare_value = self.run_compare(bench_compute_element, tested_compute_element)
            pass_status = self.check_pass(compare_value)
        else:
            logger.warning(f"not suitable for computing {self.compare_algorithm_name}, skip this.")
            compare_value = None
            pass_status = CompareConst.SKIP

        err_msg = self.err_msg_mapping.get(self.compare_algorithm_name).get(pass_status)

        compare_result = CompareResult(compare_value, pass_status, err_msg)
        return compare_result

    @staticmethod
    def convert_to_np_float64_ndarray(tensor):
        if isinstance(tensor, mindspore.Tensor):
            ndarray = tensor.astype(mindspore.float64).numpy()
        elif isinstance(tensor, torch.Tensor):
            ndarray = tensor.to(torch.float64, copy=True).numpy()
        else:
            err_msg = "BaseCompareAlgorithm.convert_to_np_float64_ndarray failed: " \
                      "input is not mindspore.Tensor or torch.Tensor"
            logger.error_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.UnsupportType))
        return ndarray

    @staticmethod
    def check_two_tensor(bench_compute_element, tested_compute_element):
        bench_parameter = bench_compute_element.get_parameter()
        tested_parameter = tested_compute_element.get_parameter()

        bench_is_tensor = isinstance(bench_parameter, (mindspore.Tensor, torch.Tensor))
        tested_is_tensor = isinstance(tested_parameter, (mindspore.Tensor, torch.Tensor))
        shape_same = bench_compute_element.get_shape() == tested_compute_element.get_shape()
        return bench_is_tensor and tested_is_tensor and shape_same

    @abstractmethod
    def check_validity(self, bench_compute_element, tested_compute_element):
        '''
        Args:
            bench_compute_element: ComputeElement
            tested_compute_element: ComputeElement

        Return:
            check_res: boolean
        '''
        raise NotImplementedError

    @abstractmethod
    def run_compare(self, bench_compute_element, tested_compute_element):
        '''
        Args:
            bench_compute_element: ComputeElement
            tested_compute_element: ComputeElement

        Return:
            compare_value: float/int
        '''
        raise NotImplementedError

    @abstractmethod
    def check_pass(self, compare_value):
        '''
        Args:
            compare_value: float/int

        Return:
            pass_status: str
        '''
        raise NotImplementedError


class CosineSimilarityCompareAlgorithm(BaseCompareAlgorithm):
    def __init__(self) -> None:
        super().__init__()
        self.compare_algorithm_name = CompareConst.COSINE

    def check_validity(self, bench_compute_element, tested_compute_element):
        return self.check_two_tensor(bench_compute_element, tested_compute_element)

    def run_compare(self, bench_compute_element, tested_compute_element):
        bench_ndarray = self.convert_to_np_float64_ndarray(bench_compute_element.get_parameter())
        tested_ndarray = self.convert_to_np_float64_ndarray(tested_compute_element.get_parameter())

        bench_norm = np.linalg.norm(bench_ndarray)
        tested_norm = np.linalg.norm(tested_ndarray)
        dot_product = np.dot(bench_ndarray.flatten(), tested_ndarray.flatten())
        cosine_similarity = (MsCompareConst.EPSILON + dot_product) / (MsCompareConst.EPSILON + bench_norm * tested_norm)
        return cosine_similarity

    def check_pass(self, compare_value):
        if compare_value > CompareConst.COS_THRESHOLD:
            return CompareConst.PASS
        else:
            return CompareConst.ERROR


class MaxAbsoluteDiffCompareAlgorithm(BaseCompareAlgorithm):
    def __init__(self) -> None:
        super().__init__()
        self.compare_algorithm_name = CompareConst.MAX_ABS_ERR

    def check_validity(self, bench_compute_element, tested_compute_element):
        return self.check_two_tensor(bench_compute_element, tested_compute_element)

    def run_compare(self, bench_compute_element, tested_compute_element):
        bench_ndarray = self.convert_to_np_float64_ndarray(bench_compute_element.get_parameter())
        tested_ndarray = self.convert_to_np_float64_ndarray(tested_compute_element.get_parameter())

        max_absolute_diff = np.max(np.abs(bench_ndarray - tested_ndarray))
        return max_absolute_diff

    def check_pass(self, compare_value):
        if compare_value < CompareConst.MAX_ABS_ERR_THRESHOLD:
            return CompareConst.PASS
        else:
            return CompareConst.ERROR


class MaxRelativeDiffCompareAlgorithm(BaseCompareAlgorithm):
    def __init__(self) -> None:
        super().__init__()
        self.compare_algorithm_name = CompareConst.MAX_RELATIVE_ERR

    def check_validity(self, bench_compute_element, tested_compute_element):
        return self.check_two_tensor(bench_compute_element, tested_compute_element)

    def run_compare(self, bench_compute_element, tested_compute_element):
        bench_ndarray = self.convert_to_np_float64_ndarray(bench_compute_element.get_parameter())
        tested_ndarray = self.convert_to_np_float64_ndarray(tested_compute_element.get_parameter())

        abs_diff = np.abs(bench_ndarray - tested_ndarray)
        bench_ndarray_nonzero = np.abs(bench_ndarray) + (bench_ndarray == 0) * MsCompareConst.EPSILON
        max_relative_diff = np.max(abs_diff / bench_ndarray_nonzero)
        return max_relative_diff

    def check_pass(self, compare_value):
        if compare_value < CompareConst.MAX_RELATIVE_ERR_THRESHOLD:
            return CompareConst.PASS
        else:
            return CompareConst.ERROR


compare_algorithms = {
    CompareConst.COSINE: CosineSimilarityCompareAlgorithm(),
    CompareConst.MAX_ABS_ERR: MaxAbsoluteDiffCompareAlgorithm(),
    CompareConst.MAX_RELATIVE_ERR: MaxRelativeDiffCompareAlgorithm(),
}



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import copy
import os
import inspect
from dataclasses import dataclass, is_dataclass
from functools import partial
from typing import Tuple, Dict, Optional, Any, Union

import numpy as np

from msprobe.core.common.const import Const
from msprobe.core.common.file_utils import save_npy
from msprobe.core.common.log import logger
from msprobe.core.common.utils import convert_tuple, CompareException


@dataclass
class ModuleForwardInputsOutputs:
    args: Optional[Tuple]
    kwargs: Optional[Dict]
    output: Any

    @property
    def args_tuple(self):
        return convert_tuple(self.args)

    @property
    def output_tuple(self):
        return convert_tuple(self.output)

    def update_output_with_args_and_kwargs(self):
        self.output = self.args + tuple(self.kwargs.values())


@dataclass
class ModuleBackwardInputsOutputs:
    grad_output: Optional[Tuple]
    grad_input: Optional[Tuple]

    @property
    def grad_input_tuple(self):
        return convert_tuple(self.grad_input)

    @property
    def grad_output_tuple(self):
        return convert_tuple(self.grad_output)


@dataclass
class ModuleBackwardInputs:
    grad_input: Optional[Tuple]

    @property
    def grad_input_tuple(self):
        return convert_tuple(self.grad_input)


@dataclass
class ModuleBackwardOutputs:
    grad_output: Optional[Tuple]

    @property
    def grad_output_tuple(self):
        return convert_tuple(self.grad_output)


class TensorStatInfo:
    def __init__(self, max_val=None, min_val=None, mean_val=None, norm_val=None):
        self.max = max_val
        self.min = min_val
        self.mean = mean_val
        self.norm = norm_val


class BaseDataProcessor:
    _recursive_key_stack = []
    builtin_type = (bool, int, float, str, slice, type(Ellipsis))
    np_type = (np.integer, np.floating, np.bool_, np.complexfloating, np.str_, np.byte, np.unicode_, np.ndarray)

    def __init__(self, config, data_writer):
        self.data_writer = data_writer
        self.config = config
        if self.data_writer is not None:
            self.data_writer.config = config
        self.api_info_struct = {}
        self.stack_info_struct = {}
        self.current_api_or_module_name = None
        self.api_data_category = None
        self.current_iter = 0
        self._return_forward_new_output = False
        self._forward_new_output = None
        self.save_name = None
        if hasattr(config, "data_mode"):
            self.allowed_data_mode = self._get_allowed_data_mode(config.data_mode)

    @property
    def data_path(self):
        return self.data_writer.dump_tensor_data_dir

    @property
    def is_terminated(self):
        return False

    @staticmethod
    def transfer_type(data):
        dtype = str(type(data))
        if 'int' in dtype:
            return int(data)
        elif 'float' in dtype:
            return float(data)
        else:
            return data

    @staticmethod
    def set_value_into_nested_structure(data_structure, indexes, value):
        '''
        Args:
            data_structure: nested data structure
            indexes: List
            value: value to be set
        '''
        if not indexes:
            raise ValueError("set_value_into_nested_structure failed: "
                             "indexes need to be non empty when set value to nested data structure")
        current_level = data_structure
        for i, index in enumerate(indexes):
            valid_for_list = isinstance(current_level, list) and isinstance(index, int) and len(current_level) > index
            valid_for_dict = isinstance(current_level, dict) and index in current_level
            is_last = i == len(indexes) - 1
            if valid_for_dict or valid_for_list:
                if is_last:
                    try:
                        current_level[index] = value
                    except Exception as e:
                        raise IndexError("set_value_into_nested_structure failed: passed indexes wrong") from e
                else:
                    try:
                        current_level = current_level[index]
                    except Exception as e:
                        raise IndexError("set_value_into_nested_structure failed: passed indexes wrong") from e
            else:
                raise ValueError("set_value_into_nested_structure failed: "
                                 "invalid data_structure type or invalid index")

    @staticmethod
    def is_distributed_op(module):
        return getattr(module, "op_is_distributed", False)

    @staticmethod
    def is_recompute(call_stack=None):
        return None

    @staticmethod
    def analyze_api_call_stack(name):
        try:
            call_stack = inspect.stack()
            if name.startswith("Primitive"):
                api_stack = call_stack[4:]
            else:
                api_stack = call_stack[5:]
        except Exception as e:
            logger.warning(f"The call stack of <{name}> failed to retrieve, {e}.")
            api_stack = None
            call_stack = None

        stack_str = []
        if api_stack:
            for (_, path, line, func, code, _) in api_stack:
                if not code:
                    continue
                if any(filter_path in path for filter_path in Const.STACK_FILTER_KEYWORDS) and \
                        Const.CALL_STACK_FLAG not in path:
                    continue
                stack_line = f"File {path}, line {str(line)}, in {func}, \n {code[0].strip()}"
                stack_str.append(stack_line)
        else:
            stack_str.append(Const.WITHOUT_CALL_STACK)
        del call_stack
        return tuple(stack_str), None

    @staticmethod
    def _analyze_builtin(arg):
        single_arg = {}
        if isinstance(arg, slice):
            # The slice parameter may be of the tensor, numpy or other types.
            # It needs to be converted to the Python value type before JSON serialization
            single_arg.update({"type": "slice"})
            values = []
            for value in [arg.start, arg.stop, arg.step]:
                if value is not None:
                    try:
                        value = int(value)
                    except ValueError:
                        logger.warning(f"The data type {type(value)} cannot be converted to int type.")
                        value = None
                values.append(value)
            single_arg.update({"value": values})
        else:
            single_arg.update({"type": type(arg).__name__})
            # When arg is Ellipsis(...) type, it needs to be converted to str("...") type
            single_arg.update({"value": arg if arg is not Ellipsis else "..."})
        return single_arg

    @staticmethod
    def _analyze_numpy(arg):
        return {"type": type(arg).__name__, "value": arg.item()}

    @staticmethod
    def _analyze_ndarray(ndarray, _):
        ndarray_json = {}
        ndarray_json.update({'type': 'numpy.ndarray'})
        ndarray_json.update({'dtype': str(ndarray.dtype)})
        ndarray_json.update({'shape': ndarray.shape})

        # 先初始化默认值
        stats = {
            "Max": None,
            "Min": None,
            "Mean": None,
            "Norm": None
        }

        try:
            # 只有非空时才尝试计算
            if ndarray.size > 0:
                stats = {
                    "Max": np.max(ndarray).item(),
                    "Min": np.min(ndarray).item(),
                    "Mean": np.mean(ndarray).item(),
                    "Norm": np.linalg.norm(ndarray).item()
                }
        except Exception as e:
            # 决定打印内容或切片
            logger.warning(f"Error analyzing ndarray stats: {e}")

        # 最后一次性更新
        ndarray_json.update(stats)

        return ndarray_json

    @staticmethod
    def _get_allowed_data_mode(data_mode):
        if Const.ALL in data_mode:
            allowed_data_mode = [Const.FORWARD, Const.BACKWARD, Const.INPUT, Const.OUTPUT]
        else:
            allowed_data_mode = list(set(data_mode))
            if Const.FORWARD not in allowed_data_mode and Const.BACKWARD not in allowed_data_mode:
                allowed_data_mode += [Const.FORWARD, Const.BACKWARD]
            if Const.INPUT not in allowed_data_mode and Const.OUTPUT not in allowed_data_mode:
                allowed_data_mode += [Const.INPUT, Const.OUTPUT]
        return allowed_data_mode

    @classmethod
    def get_special_types(cls):
        return cls.builtin_type + cls.np_type

    @classmethod
    def recursive_apply_transform(cls, args, transform, depth=0) -> Union[dict, list, None]:
        if depth > Const.DUMP_MAX_DEPTH:
            logger.error(f"The maximum depth of recursive transform, {Const.DUMP_MAX_DEPTH} is reached.")
            raise CompareException(CompareException.RECURSION_LIMIT_ERROR)
        if isinstance(args, cls.get_special_types()):
            arg_transform = transform(args, cls._recursive_key_stack)
            return arg_transform
        elif isinstance(args, tuple) and hasattr(args, '_fields'):
            # namedtuple to dict
            args_dict = {field: getattr(args, field) for field in args._fields}
            return cls.apply_transform_dict(args_dict, transform, depth)
        elif is_dataclass(args):
            # dataclass to dict
            args_dict = {field: getattr(args, field) for field in args.__dataclass_fields__}
            return cls.apply_transform_dict(args_dict, transform, depth)
        elif isinstance(args, (list, tuple)):
            result_list = cls.apply_transform_list(args, transform, depth)
            return result_list
        elif isinstance(args, dict):
            return cls.apply_transform_dict(args, transform, depth)
        elif args is not None:
            logger.debug(f"Data type {type(args)} is not supported.")
            return None
        else:
            return None

    @classmethod
    def apply_transform_dict(cls, args, transform, depth):
        result_dict = {}
        for k, arg in args.items():
            cls._recursive_key_stack.append(k)
            result_dict[k] = cls.recursive_apply_transform(arg, transform, depth=depth + 1)
            cls._recursive_key_stack.pop()
        return result_dict

    @classmethod
    def apply_transform_list(cls, args, transform, depth):
        result_list = []
        for i, arg in enumerate(args):
            cls._recursive_key_stack.append(i)
            result_list.append(cls.recursive_apply_transform(arg, transform, depth=depth + 1))
            cls._recursive_key_stack.pop()
        return result_list

    @classmethod
    def register_hook_single_element(cls, element, suffix_stack, hook_fn):
        if cls.is_hookable_element(element):
            indexes = copy.deepcopy(suffix_stack)
            wrap_hook_fn = partial(hook_fn, indexes=indexes)

            def real_hook_fn(grad):
                return wrap_hook_fn(grad)

            element.register_hook(real_hook_fn)

    def if_return_forward_new_output(self):
        return self._return_forward_new_output

    def get_forward_new_output(self):
        self._return_forward_new_output = False
        return self._forward_new_output

    def update_iter(self, current_iter):
        self.current_iter = current_iter

    def update_api_or_module_name(self, api_or_module_name):
        if self.current_api_or_module_name != api_or_module_name:
            self.current_api_or_module_name = api_or_module_name

    def is_dump_for_data_mode(self, forward_backward, input_output):
        """
        Compare the parameters with data_mode to determine whether to dump.

        Args:
            forward_backward(str): The forward or backward mode to check.
            input_output(str): The input or output mode to check.

        Return:
            bool: True if the parameters are in data_mode or data_mode is all, False otherwise.
        """
        return forward_backward in self.allowed_data_mode and input_output in self.allowed_data_mode

    def analyze_element(self, element):
        return self.recursive_apply_transform(element, self.analyze_single_element)

    def analyze_forward_input(self, name, module, module_input_output: ModuleForwardInputsOutputs):
        api_info_struct = {}
        # check whether data_mode contains forward or input
        if self.is_dump_for_data_mode(Const.FORWARD, Const.INPUT):
            api_info_struct[name] = {}
            self.api_data_category = Const.INPUT
            args_info_list = self.analyze_element(module_input_output.args_tuple)
            api_info_struct[name][Const.INPUT_ARGS] = args_info_list
            self.api_data_category = Const.KWARGS
            kwargs_info_list = self.analyze_element(module_input_output.kwargs)
            api_info_struct[name][Const.INPUT_KWARGS] = kwargs_info_list

        return api_info_struct

    def analyze_forward_output(self, name, module, module_input_output: ModuleForwardInputsOutputs):
        if self.is_distributed_op(module):
            module_input_output.update_output_with_args_and_kwargs()
        api_info_struct = {}
        # check whether data_mode contains forward or input
        if self.is_dump_for_data_mode(Const.FORWARD, Const.OUTPUT):
            api_info_struct[name] = {}
            self.api_data_category = Const.OUTPUT
            output_info_list = self.analyze_element(module_input_output.output_tuple)
            api_info_struct[name][Const.OUTPUT] = output_info_list

        return api_info_struct

    def analyze_forward(self, name, module, module_input_output: ModuleForwardInputsOutputs):
        api_info_struct = {}
        # check whether data_mode contains forward or input
        if self.is_dump_for_data_mode(Const.FORWARD, Const.INPUT):
            api_info_struct[name] = {}
            self.api_data_category = Const.INPUT
            args_info_list = self.analyze_element(module_input_output.args_tuple)
            api_info_struct[name][Const.INPUT_ARGS] = args_info_list
            self.api_data_category = Const.KWARGS
            kwargs_info_list = self.analyze_element(module_input_output.kwargs)
            api_info_struct[name][Const.INPUT_KWARGS] = kwargs_info_list

        # check whether data_mode contains forward or output
        if self.is_dump_for_data_mode(Const.FORWARD, Const.OUTPUT):
            api_info_struct[name] = api_info_struct.get(name, {})
            self.api_data_category = Const.OUTPUT
            output_info_list = self.analyze_element(module_input_output.output_tuple)
            api_info_struct[name][Const.OUTPUT] = output_info_list

        if name in api_info_struct and hasattr(module_input_output, Const.PARAMS):
            self.api_data_category = Const.PARAMS
            api_info_struct[name][Const.PARAMS] = self.analyze_element(getattr(module_input_output, Const.PARAMS))

        return api_info_struct

    def analyze_backward(self, name, module, module_input_output: ModuleBackwardInputsOutputs):
        api_info_struct = {}
        if self.is_dump_for_data_mode(Const.BACKWARD, Const.INPUT):
            api_info_struct[name] = {}
            self.api_data_category = Const.INPUT
            input_info_list = self.analyze_element(module_input_output.grad_input_tuple)
            api_info_struct[name][Const.INPUT] = input_info_list

        if self.is_dump_for_data_mode(Const.BACKWARD, Const.OUTPUT):
            api_info_struct[name] = api_info_struct.get(name, {})
            self.api_data_category = Const.OUTPUT
            output_info_list = self.analyze_element(module_input_output.grad_output_tuple)
            api_info_struct[name][Const.OUTPUT] = output_info_list

        return api_info_struct

    def analyze_backward_input(self, name, module,
                               module_input_output: ModuleBackwardInputs):
        api_info_struct = {}
        if self.is_dump_for_data_mode(Const.BACKWARD, Const.INPUT):
            api_info_struct[name] = {}
            self.api_data_category = Const.INPUT

            input_info_list = self.analyze_element(module_input_output.grad_input_tuple)
            api_info_struct[name][Const.INPUT] = input_info_list
        return api_info_struct

    def analyze_backward_output(self, name, module,
                                module_input_output: ModuleBackwardOutputs):
        api_info_struct = {}
        if self.is_dump_for_data_mode(Const.BACKWARD, Const.OUTPUT):
            api_info_struct[name] = {}
            self.api_data_category = Const.OUTPUT

            output_info_list = self.analyze_element(module_input_output.grad_output_tuple)
            api_info_struct[name][Const.OUTPUT] = output_info_list
        return api_info_struct

    def analyze_params(self, name, param_name, grad):
        api_info_struct = {}
        self.save_name = name + Const.SEP + param_name
        data_info = self.analyze_element(grad)
        self.save_name = None
        grad_info_dict = {param_name: [data_info]}
        api_info_struct[name] = grad_info_dict
        return api_info_struct

    def get_save_file_path(self, suffix):
        file_format = Const.PT_SUFFIX if self.config.framework == Const.PT_FRAMEWORK else Const.NUMPY_SUFFIX
        if self.save_name is not None:
            dump_data_name = (self.save_name + file_format)
        else:
            suffix_with_seq = (Const.SEP + suffix) if suffix else ""
            dump_data_name = (self.current_api_or_module_name + Const.SEP + self.api_data_category + suffix_with_seq +
                              file_format)
        file_path = os.path.join(self.data_writer.dump_tensor_data_dir, dump_data_name)
        return dump_data_name, file_path

    def analyze_element_to_all_none(self, element):
        return self.recursive_apply_transform(element, lambda element, stack: None)

    def analyze_debug_forward(self, variable, name_with_count):
        self.current_api_or_module_name = name_with_count
        self.api_data_category = Const.DEBUG
        # these two attributes are used to construct tensor file name {name_with_count}.debug.{indexes}.npy/pt
        data_info = self.analyze_element(variable)
        return data_info

    def analyze_debug_backward(self, variable, grad_name_with_count_category, nested_data_structure):
        def hook_fn(grad, indexes):
            suffix = Const.SEP.join([str(index) for index in indexes])
            suffix_with_sep = (Const.SEP + suffix) if suffix else ""
            self.save_name = grad_name_with_count_category + suffix_with_sep
            grad_data_info = self.analyze_element(grad)
            self.save_name = None
            full_index = [grad_name_with_count_category] + indexes
            try:
                self.set_value_into_nested_structure(nested_data_structure, full_index, grad_data_info)
            except (ValueError, IndexError) as e:
                logger.warning(f"error occurred while recording statistics of {grad_name_with_count_category} variable,"
                               f"skip current recording, detailed information: {e}")
            return grad

        wrap_register_hook_single_element = partial(self.register_hook_single_element, hook_fn=hook_fn)
        self.recursive_apply_transform(variable, wrap_register_hook_single_element)

    def _analyze_and_save_ndarray(self, ndarray, suffix):
        dump_data_name, file_path = self.get_save_file_path(suffix)
        save_npy(ndarray, file_path)
        ndarray_json = BaseDataProcessor._analyze_ndarray(ndarray, suffix)
        ndarray_json.update({"data_name": dump_data_name})
        return ndarray_json



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


import random
from functools import wraps
from typing import Callable, List, Dict, Tuple, Optional
import inspect
import os
import json
from collections import defaultdict
import difflib

import numpy as np
import pandas as pd
from msprobe.core.config_check.config_checker import register_checker_item, register_pre_forward_fun_list
from msprobe.core.common.file_utils import create_file_in_zip, load_json
from msprobe.core.config_check.checkers.base_checker import BaseChecker
from msprobe.core.config_check.utils.utils import config_checking_print
from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.common.const import Const
from msprobe.core.common.log import logger


# 数据结构：{随机操作名字: [{count: 调用次数, stack: 调用栈列表}]}
random_op_stats = defaultdict(list)


def get_call_stack(frame) -> List[str]:
    """获取详细的调用栈信息，每个元素包含完整路径、行号、函数名和代码行"""
    stack = []
    current_frame = frame.f_back  # 跳过当前函数
    
    while current_frame:
        frame_info = inspect.getframeinfo(current_frame)
        filename = os.path.abspath(frame_info.filename)
        code_line = frame_info.code_context[0].strip() if frame_info.code_context else ""
        
        # 格式化为详细的栈帧信息
        stack_entry = f"File {filename}, line {frame_info.lineno}, in {frame_info.function}, {code_line}"
        stack.append(stack_entry)
        
        current_frame = current_frame.f_back
    
    # 反转堆栈以显示正确的调用顺序（栈底到栈顶）
    return stack[::-1]


def track_random_call(func: Callable, name: str):
    """记录随机函数的调用信息"""
    @wraps(func)
    def wrapper(*args, **kwargs):
        frame = inspect.currentframe()
        stack = get_call_stack(frame)
        
        # 更新调用统计：操作名 -> [{count: 次数, stack: 调用栈列表}]
        # 检查是否已有相同调用栈的记录
        for entry in random_op_stats[name]:
            if entry['stack'] == stack:
                entry['count'] += 1
                break
        else:
            # 新增调用栈记录
            random_op_stats[name].append({'count': 1, 'stack': stack})
        
        try:
            result = func(*args, **kwargs)
            return result
        except Exception as e:
            raise e
        finally:
            del frame
            
    return wrapper


def load_stats_files(directory: str) -> Dict[str, Dict[str, List[Dict]]]:
    """加载目录下所有统计文件并按rank组织数据"""
    rank_data = {}
    for file in os.listdir(directory):
        file_path = os.path.join(directory, file)
        if file.startswith('rank') and file.endswith('.json'):
            rank = os.path.basename(file.split('.')[0])[4:]
            if not rank or not rank.isdigit():
                logger.error(f"extract rank id from {file} failed")
                raise ValueError
            
            # 加载并存储数据
            data = load_json(file_path)
            rank_data[int(rank)] = data

    return rank_data


def stack_match(stack1: List[str], stack2: List[str], threshold: float = 0.8) -> bool:
    """
    比较两个调用栈是否相似，同时考虑路径、函数名和代码行（各占1/3），每一层的相似度阈值需要达到0.8
    
    参数:
    - stack1: 第一个调用栈列表
    - stack2: 第二个调用栈列表
    - threshold: 相似度阈值，默认0.8
    
    返回:
    - 两个调用栈是否相似的布尔值
    """
    if len(stack1) != len(stack2):
        return False
    
    for frame1, frame2 in zip(stack1, stack2):
        # 提取路径、函数名和代码行
        path1, func1, code1 = _parse_frame(frame1)
        path2, func2, code2 = _parse_frame(frame2)
        
        # 计算相似度得分 (路径、函数名、代码行各占1/3权重)
        path_score = _compare_path(path1, path2)
        func_score = 1.0 if func1 == func2 else 0.0
        # 代码相似度
        code_score = difflib.SequenceMatcher(None, code1, code2).ratio()
        
        frame_score = (path_score + func_score + code_score) / 3.0
        if frame_score < threshold:
            return False

    return True


def _parse_frame(frame: str) -> Tuple[str, str, str]:
    """
    解析栈帧字符串，提取路径、函数名和代码行

    参数:
    - frame: 栈帧字符串。格式为"File {path}, line {line}, in {func}, {code}"
    
    返回:
    - path, func, code
    """
    path = func = code = ''
    stack_info = frame.split(' ')
    if len(stack_info) > 6:
        path = stack_info[1][:-1]
        func = stack_info[5][:-1]
        code = ' '.join(stack_info[6:])
    return path, func, code


def _compare_path(path1: str, path2: str) -> float:
    """比较两个路径的相似度，只考虑文件名"""
    if not path1 or not path2:
        return 0.0
    
    # 提取文件名（忽略目录路径）
    file1 = os.path.basename(path1)
    file2 = os.path.basename(path2)
    
    return 1.0 if file1 == file2 else 0.0


def find_matching_stack(bench_stack: List[str], cmp_stacks: List[Dict]) -> Optional[Dict]:
    """
    查找匹配的调用栈
    
    参数:
    - bench_stack: 基准侧的调用栈列表
    - cmp_stacks: 比较侧的调用栈条目列表，每个条目是{'count': 次数, 'stack': 调用栈列表}
    
    返回:
    - 匹配的调用栈条目或None
    """
    for cmp_entry in cmp_stacks:
        if stack_match(cmp_entry['stack'], bench_stack):
            return cmp_entry
    
    return None


def stack_list_to_string(stack_list):
    """
    将调用栈列表转换为换行分隔的字符串
    如果输入是特殊标记（如"no match stack"），则直接返回
    """
    if isinstance(stack_list, list):
        return '\n'.join(stack_list)
    return stack_list


def compare_random_calls(bench_dir: str = 'bench', cmp_dir: str = 'cmp') -> pd.DataFrame:
    """比较两个目录下的随机调用栈统计，生成详细比对结果"""
    bench_rank_data = load_stats_files(bench_dir)
    cmp_rank_data = load_stats_files(cmp_dir)
    
    # 获取所有rank
    all_ranks = sorted(set(bench_rank_data.keys()) | set(cmp_rank_data.keys()))
    
    results = []
    
    for rank in all_ranks:
        bench_data = bench_rank_data.get(rank, {})
        cmp_data = cmp_rank_data.get(rank, {})
        
        # 获取所有操作
        all_ops = set(bench_data.keys()) | set(cmp_data.keys())
        
        for op in all_ops:
            bench_stacks = bench_data.get(op, [])
            cmp_stacks = cmp_data.get(op, [])
            
            # 处理bench侧的每个调用栈
            for bench_entry in bench_stacks:
                bench_stack = bench_entry['stack']
                bench_count = bench_entry['count']
                
                # 查找匹配的cmp侧调用栈
                cmp_entry = find_matching_stack(bench_stack, cmp_stacks)
                
                if cmp_entry:
                    cmp_count = cmp_entry['count']
                    check_result = bench_count == cmp_count
                    results.append([op, rank, bench_stack, cmp_entry['stack'], bench_count, cmp_count, check_result])
                else:
                    # 没有匹配的调用栈
                    results.append([op, rank, bench_stack, "no match stack", bench_count, 0, False])
            
            # 处理cmp侧中没有在bench侧出现的调用栈
            for cmp_entry in cmp_stacks:
                cmp_stack = cmp_entry['stack']
                # 检查是否已经在上面处理过
                if not any(stack_match(bench_entry['stack'], cmp_stack) for bench_entry in bench_stacks):
                    results.append([op, rank, "no match stack", cmp_stack, 0, cmp_entry['count'], False])
    
    # 创建DataFrame
    df = pd.DataFrame(results, columns=RandomChecker.result_header)
    
    # 应用转换函数
    df['bench_stack'] = df['bench_stack'].apply(stack_list_to_string)
    df['cmp_stack'] = df['cmp_stack'].apply(stack_list_to_string)
    
    return df


def torch_patchs():
    """补丁Torch随机函数"""
    import torch
    torch_patches = {
        'rand': torch.rand,
        'randint': torch.randint,
        'randn': torch.randn,
        'rand_like': torch.rand_like,
        'randint_like': torch.randint_like,
        'randn_like': torch.randn_like,
        'manual_seed': torch.manual_seed
    }
    for name, func in torch_patches.items():
        setattr(torch, name, track_random_call(func, f"torch.{name}"))

    tensor_patches = {
        'exponential_': torch.Tensor.exponential_,
        'geometric_': torch.Tensor.geometric_,
        'log_normal_': torch.Tensor.log_normal_,
        'cauchy_': torch.Tensor.cauchy_
    }
    for name, func in tensor_patches.items():
        setattr(torch.Tensor, name, track_random_call(func, f"torch.Tensor.{name}"))


def mindspore_patchs():
    """补丁MindSpore随机函数"""
    import mindspore

    mindspore_ops_patches = {
        'rand': mindspore.ops.rand,
        'randint': mindspore.ops.randint,
        'randn': mindspore.ops.randn
    }
    for name, func in mindspore_ops_patches.items():
        setattr(mindspore.ops, name, track_random_call(func, f"mindspore.ops.{name}"))

    mindspore_patches = {
        'manual_seed': mindspore.set_seed
    }
    for name, func in mindspore_patches.items():
        setattr(mindspore, name, track_random_call(func, f"mindspore.{name}"))
    

@register_checker_item("random")
class RandomChecker(BaseChecker):
    input_needed = None
    target_name_in_zip = "random"
    result_header = ['op', 'rank', 'bench_stack', 'cmp_stack', 'bench_count', 'cmp_count', 'check_result']
    write_once = False

    @staticmethod
    def pack(pack_input):
        """打包随机调用统计到zip文件"""
        output_zip_path = pack_input.output_zip_path
        
        def collect_input(model, args, kwargs, step):
            if RandomChecker.write_once:
                return
                
            random_stats_dir = os.path.join(RandomChecker.target_name_in_zip)
            stats_filepath = os.path.join(random_stats_dir, f"rank{FmkAdp.get_rank_id()}.json")
            
            # 转换为JSON格式：{操作名: [{count: 次数, stack: 调用栈列表}]}
            stats_json = {}
            for op_name, entries in random_op_stats.items():
                stats_json[op_name] = entries
            
            create_file_in_zip(output_zip_path, stats_filepath, json.dumps(stats_json, indent=4))
            config_checking_print(f"已将随机调用统计打包到: {stats_filepath}")
            RandomChecker.write_once = True
            
        register_pre_forward_fun_list(collect_input)

    @staticmethod
    def compare(bench_dir, cmp_dir, output_path):
        """比较两组随机调用统计"""
        bench_stats_path = os.path.join(bench_dir, RandomChecker.target_name_in_zip)
        cmp_stats_path = os.path.join(cmp_dir, RandomChecker.target_name_in_zip)
        
        df = compare_random_calls(bench_stats_path, cmp_stats_path)
        pass_check = Const.CONFIG_CHECK_PASS if False not in df['check_result'].values else Const.CONFIG_CHECK_ERROR
        
        return RandomChecker.target_name_in_zip, pass_check, df

    @staticmethod
    def apply_patches(fmk=Const.PT_FRAMEWORK):
        """应用随机函数补丁"""
        # 补丁Python random模块
        random_patches = {
            'random': random.random,
            'randint': random.randint,
            'uniform': random.uniform,
            'choice': random.choice
        }
        for name, func in random_patches.items():
            setattr(random, name, track_random_call(func, f"random.{name}"))

        # 补丁Numpy随机函数
        np_random_patches = {
            'rand': np.random.rand,
            'randint': np.random.randint,
            'choice': np.random.choice,
            'normal': np.random.normal
        }
        for name, func in np_random_patches.items():
            setattr(np.random, name, track_random_call(func, f"np.random.{name}"))

        # 补丁框架特定随机函数
        if fmk == Const.PT_FRAMEWORK:
            torch_patchs()
        elif fmk == Const.MS_FRAMEWORK:
            mindspore_patchs()
        else:
            raise Exception(f"不支持的框架: {fmk}, 支持的框架: {FmkAdp.supported_fmk}")



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the common function.
"""
import re
import os

from functools import wraps
import numpy as np
from msprobe.msaccucmp.dump_parse.proto_dump_data import DumpData

from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.reg_manager import RegManager
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.dump_parse.big_dump_data import DumpDataHandler
from msprobe.msaccucmp.dump_parse.dump_data_object import DumpDataObj, DumpTensor
from msprobe.msaccucmp.dump_parse.nano_dump_data import NanoDumpData, NanoDumpDataParser, NanoDumpDataHandler


class SortMode:
    """
    The class of sort mode
    """
    hash_to_file_name_map = {}

    def __init__(self, parameter):
        self.parameter = parameter

    def __call__(self: any, wrap_function):
        """
        the wrapper of get info to sort
        @param wrap_function: file name
        @return: Basis of sorted
        """
        @wraps(wrap_function)
        def inner(*args, **kwargs):
            file_path = wrap_function(*args, **kwargs)
            file_name = os.path.basename(file_path)
            file_name = self.hash_to_file_name_map.get(file_name) if file_name.isdigit() else file_name
            if file_name is None:
                log.print_warn_log('The file_name is invalid, failed to sort')
                return ConstManager.INVALID_SORT_MODE
            file_split = file_name.split('.')
            if self.parameter == ConstManager.NORMAL_MODE or \
                    self.parameter == ConstManager.FFTS_TIMESTAMP:
                return self._parameter_timestamp(file_split, file_name)
            elif self.parameter == ConstManager.AUTOMATIC_MODE:
                return self._parameter_auto(file_split, file_name)
            elif self.parameter == ConstManager.MANUAL_MODE:
                return self._parameter_manual(file_split, file_name)
            else:
                log.print_warn_log('The sort mode parameter is invalid, failed to sort')
                return ConstManager.INVALID_SORT_MODE
        return inner

    @staticmethod
    def _parameter_manual(file_split, file_name):
        # Conv2D.partition0_rank2_new_sub_graph15_sgt_graph_0_fp32_vars_conv2d_39_Conv2D_lxslice0. \
        # 2.9.1670205071724946.4.487.0.0
        slice_x = file_split[1][-1]
        if not slice_x.isdigit():
            log.print_warn_log(
                'The file name \"{}\"\'s slice_x is invalid.'.format(file_name))
            return ConstManager.INVALID_SLICE_X
        return int(slice_x)

    @staticmethod
    def _parameter_auto(file_split, file_name):
        thread_id = file_split[-2]
        if not thread_id.isdigit():
            log.print_warn_log(
                'The file name \"{}\"\'s thread_id is invalid.'.format(file_name))
            return ConstManager.INVALID_THREAD_ID
        return int(thread_id)

    def _parameter_timestamp(self, file_split, file_name):
        if self.parameter == ConstManager.FFTS_TIMESTAMP:
            timestamp = file_split[4]
        elif file_name.endswith(
                (ConstManager.STANDARD_SUFFIX, ConstManager.NUMPY_SUFFIX, ConstManager.QUANT_SUFFIX)):
            timestamp = file_split[2]
        else:
            timestamp = file_split[-1]
        if not check_valid_timestamp(timestamp):
            log.print_warn_log(
                'The file name \"{}\"\'s timestamp is invalid.'.format(file_name))
            return ConstManager.INVALID_TIMESTAMP
        return int(timestamp)


@SortMode(ConstManager.AUTOMATIC_MODE)
def get_ffts_auto(file_name):
    """
    get thread id of ffts auto mode from file name
    @param file_name: file name
    @return: thread id
    """
    return file_name


@SortMode(ConstManager.MANUAL_MODE)
def get_ffts_manual(file_name):
    """
    get slice X of ffts manual mode from file name
    @param file_name: file name
    @return: slice X
    """
    return file_name


@SortMode(ConstManager.NORMAL_MODE)
def get_normal_timestamp(file_name):
    """
    get timestamp of normal mode
    @param file_name: file name
    @return: timestamp
    """
    return file_name


@SortMode(ConstManager.FFTS_TIMESTAMP)
def get_ffts_timestamp(file_name):
    """
    get timestamp of ffts mode
    @param file_name: file name
    @return: timestamp
    """
    return file_name


def sort_dump_file_list(dump_file_type: int, dump_file_list: list) -> list:
    """
    sort dump file list by different dump mode
    @param dump_file_type: dump data mode
    @param dump_file_list: dump file list
    @return: sorted dump file list
    """
    if dump_file_type == ConstManager.NORMAL_MODE:
        dump_file_list.sort(key=get_normal_timestamp)
    elif dump_file_type == ConstManager.AUTOMATIC_MODE or dump_file_type == ConstManager.MANUAL_MODE:
        dump_file_list.sort(key=get_ffts_timestamp)
        if dump_file_type == ConstManager.AUTOMATIC_MODE:
            dump_file_list.sort(key=get_ffts_auto)
        elif dump_file_type == ConstManager.MANUAL_MODE:
            dump_file_list.sort(key=get_ffts_manual)
    return dump_file_list


def read_numpy_file(path: str) -> any:
    """
    Read numpy file
    :param path: the numpy file path
    :return: numpy data
    """
    return DumpDataHandler(path).read_numpy_file()


def convert_dump_data_object(input_path, dump_version) -> DumpDataObj:
    """
    This is a wrapper
    @param wrap_function: function need to be wrapped
    @return: inner function
    """
    try:
        dump_data = DumpDataHandler(input_path).parse_dump_data()
    except CompareError as error:
        if error.code == CompareError.MSACCUCMP_UNMATCH_STANDARD_DUMP_SIZE:
            dump_data = DumpData()
        else:
            raise error
    dump_data_object = convert_dump_data(dump_data)
    return dump_data_object


def convert_dump_data(dump_data: DumpData) -> DumpDataObj:
    """
    Convert dump_data to DumpDataObj
    @param dump_data: DumpData object
    @return: DumpDataObj object
    """
    dump_data_object = DumpDataObj(dump_data)

    return dump_data_object


def convert_nano_dump_data(nano_dump_data: NanoDumpData) -> DumpDataObj:
    """
    Convert dump_data to DumpDataObj
    @param dump_data: DumpData object
    @return: DumpDataObj object
    """
    dump_data_object = DumpDataObj(dump_data=None, nano_dump_data=nano_dump_data)
    return dump_data_object


def parse_dump_file(input_path: str, dump_version: int) -> DumpDataObj:
    """
    Parse dump file
    :param input_path: the input file path
    :param dump_version: the dump version
    :return: DumpData
    """
    if NanoDumpDataHandler(input_path).check_is_nano_dump_format():
        nano_dump_data = NanoDumpDataHandler(input_path).parse_dump_data()
        return convert_nano_dump_data(nano_dump_data)
    else:
        return convert_dump_data_object(input_path, dump_version)


def get_op_type_from_file_name(dump_path: str):
    """
    get op_type from dump file name
    """
    dump_file_name = os.path.basename(dump_path).replace("*", "0")
    is_match, match = RegManager.match_group(RegManager.OFFLINE_DUMP_PATTERN, dump_file_name)
    if is_match:
        op_type_end_index = dump_file_name.find('.')
        return dump_file_name[:op_type_end_index]
    return ConstManager.NAN


def check_valid_timestamp(timestamp) -> bool:
    """
    Check if timestamp format is valid
    @param timestamp: timestamp from dump_file_path
    @return: True or False
    """
    return len(timestamp) == ConstManager.TIMESTAMP_LENGTH and timestamp.isdigit()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple

from msprobe.core.common.const import Const, MonitorConst
from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.common.log import logger
from msprobe.core.monitor_v2.base import BaseMonitorV2
from msprobe.core.monitor_v2.utils import get_vpp_stage_from_tag, iter_model_modules


def _iter_named_modules(model: Any):
    return iter_model_modules(model)


class ModuleMonitorV2(BaseMonitorV2, ABC):
    def __init__(self) -> None:
        super().__init__()
        self._model: Optional[Any] = None
        self._hooks: List[Any] = []
        self._targets: List[str] = []
        self._eps: float = 1e-8

    @staticmethod
    def _try_register_hook(register_fn: Any, hook: Any) -> Any:
        if register_fn is None:
            return None
        try:
            return register_fn(hook, with_kwargs=True)
        except TypeError:
            return register_fn(hook)

    def set_config(self, config: Dict[str, Any]) -> None:
        super().set_config(config)
        self.config = config
        self._targets = config.get("targets", [])
        self._eps = float(config.get("eps", 1e-8))

    @abstractmethod
    def _register_backward_hook(self, module: Any, module_name: str) -> Any:
        ...

    @abstractmethod
    def _compute_metrics(self, tag2tensor: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        ...

    def start(self, model: Any = None, optimizer: Any = None, **context: Any) -> None:
        self.set_context(**context)
        if self._model is not None:
            return
        model = model or self._context.get("model")
        if model is None:
            raise ValueError("Model must be provided to start ModuleMonitorV2")
        self._model = model
        self._register_hooks(model)

    def stop(self) -> None:
        if self._model is None:
            return
        for hook in self._hooks:
            try:
                hook.remove()
            except Exception as exc:
                logger.warning(f"[monitor_v2] Failed to remove module hook: {exc}")
        self._hooks.clear()
        self._model = None

    def _register_hooks(self, model: Any) -> None:
        targets = self._targets
        for name, module in _iter_named_modules(model):
            if not targets or any(keyword in name for keyword in targets):
                self._register_module_hooks(module, name)

    def _register_module_hooks(self, module: Any, module_name: str) -> None:
        if not FmkAdp.is_nn_module(module):
            return
        self._add_hook(self._try_register_hook(
            module.register_forward_pre_hook,
            self._build_forward_pre_hook(module_name),
        ))
        self._add_hook(self._try_register_hook(
            module.register_forward_hook,
            self._build_forward_hook(module_name),
        ))
        backward_handle = self._register_backward_hook(module, module_name)
        self._add_hook(backward_handle)

    def _build_forward_pre_hook(self, module_name: str):
        def hook(module: Any, inputs: Any, kwargs: Optional[Dict[str, Any]] = None):
            tag2tensor, idx = self._collect_tagged_tensors(module_name, "input", inputs)
            if kwargs:
                extras, _ = self._collect_tagged_tensors(module_name, "input", tuple(kwargs.values()), idx)
                tag2tensor.update(extras)
            self._record_hook_data(module_name, "input", tag2tensor, "forward")
        return hook

    def _build_forward_hook(self, module_name: str):
        def hook(module: Any, *hook_args: Any, **hook_kwargs: Any):
            if len(hook_args) < 2:
                return
            outputs = hook_args[-1]
            tag2tensor, _ = self._collect_tagged_tensors(module_name, "output", outputs)
            self._record_hook_data(module_name, "output", tag2tensor, "forward")
        return hook

    def _build_backward_hook(self, module_name: str):
        def hook(module: Any, grad_inputs: Any, grad_outputs: Any):
            input_tags, _ = self._collect_tagged_tensors(module_name, "grad_input", grad_inputs)
            output_tags, _ = self._collect_tagged_tensors(module_name, "grad_output", grad_outputs)
            self._record_hook_data(module_name, "grad_input", input_tags, "backward")
            self._record_hook_data(module_name, "grad_output", output_tags, "backward")
        return hook

    def _record_hook_data(self, module_name: str, io_kind: str, tag2tensor: Dict[str, Any], hook_type: str) -> None:
        if not tag2tensor:
            return
        stats = self._compute_metrics(tag2tensor)
        if not stats:
            return
        for tag, op_dict in stats.items():
            vpp_stage = get_vpp_stage_from_tag(tag)
            row = {
                "module_name": tag,
                "scope": hook_type,
                "stats": {op: self._detach_stat(value) for op, value in op_dict.items()},
            }
            if vpp_stage is not None:
                row["vpp_stage"] = vpp_stage
            else:
                row["vpp_stage"] = MonitorConst.DEFAULT_STAGE
            self._rows.append(row)

    def _collect_tagged_tensors(
        self,
        module_name: str,
        io_kind: str,
        tensors: Any,
        start_idx: int = 0,
    ) -> Tuple[Dict[str, Any], int]:
        tagged: Dict[str, Any] = {}
        idx = start_idx
        if tensors is None:
            return tagged, idx
        if FmkAdp.is_tensor(tensors):
            tagged[f"{module_name}.{io_kind}.{idx}"] = tensors
            return tagged, idx + 1
        if isinstance(tensors, (list, tuple)):
            for tensor in tensors:
                if not FmkAdp.is_tensor(tensor):
                    continue
                tagged[f"{module_name}.{io_kind}.{idx}"] = tensor
                idx += 1
        return tagged, idx

    def _detach_stat(self, value: Any) -> Any:
        if value is None or not FmkAdp.is_tensor(value):
            return value
        if FmkAdp.fmk == Const.PT_FRAMEWORK:
            return value.detach()
        return value

    def _add_hook(self, handle: Any) -> None:
        if handle is not None:
            self._hooks.append(handle)


class PyTorchModuleMonitorV2(ModuleMonitorV2):
    def _register_backward_hook(self, module: Any, module_name: str) -> Any:
        return module.register_full_backward_hook(self._build_backward_hook(module_name))

    def _compute_metrics(self, tag2tensor: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        from msprobe.pytorch.monitor.module_metric import get_metrics

        return get_metrics(self._ops, tag2tensor, self._eps, {})


class MindSporeModuleMonitorV2(ModuleMonitorV2):
    def _register_backward_hook(self, module: Any, module_name: str) -> Any:
        return module.register_backward_hook(self._build_backward_hook(module_name))

    def _compute_metrics(self, tag2tensor: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        from msprobe.mindspore.monitor.utils import get_metrics

        return get_metrics(self._ops, tag2tensor, self._eps, {})



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import sqlite3
import json
import re
import time
from typing import Dict
from msprobe.core.common.log import logger
from msprobe.core.common.file_utils import change_mode, check_path_before_create, FileChecker
from msprobe.core.common.const import FileCheckConst, Const
from msprobe.visualization.utils import GraphConst, update_shared_dict, update_pbar_info, post_process_db_pbar
from msprobe.visualization.builder.msprobe_adapter import format_node_data

TEXT_PRIMARY_KEY = 'TEXT PRIMARY KEY'
TEXT_NOT_NULL = 'TEXT NOT NULL'
INTEGER_NOT_NULL = 'INTEGER NOT NULL'
TEXT = 'TEXT'
INTEGER = 'INTEGER'
DB_INSERT_SIZE = 1000

node_columns = {
    'id': TEXT_PRIMARY_KEY,
    'graph_id': TEXT_NOT_NULL,
    'node_order': INTEGER_NOT_NULL,
    'node_name': TEXT_NOT_NULL,
    'node_type': TEXT_NOT_NULL,
    'up_node': TEXT,
    'sub_nodes': TEXT,
    'precision_index': INTEGER,
    'overflow_level': TEXT,
    'micro_step_id': INTEGER_NOT_NULL,
    'matched_node_link': TEXT,
    'stack_id': TEXT,
    'parallel_merge_info': TEXT,
    'matched_distributed': TEXT,
    'modified': INTEGER_NOT_NULL,
    'input_data': TEXT,
    'output_data': TEXT,
    'data_source': TEXT,
    'dump_data_dir': TEXT,
    'step': INTEGER_NOT_NULL,
    'rank': INTEGER_NOT_NULL,
    'is_distributed': INTEGER
}

config_columns = {
    'id': TEXT_PRIMARY_KEY,
    'graph_type': TEXT_NOT_NULL,
    'task': TEXT,
    'tool_tip': TEXT,
    'micro_steps': INTEGER,
    'overflow_check': INTEGER,
    'node_colors': TEXT_NOT_NULL,
    'rank_list': TEXT_NOT_NULL,
    'step_list': TEXT_NOT_NULL
}

stack_columns = {
    'id': TEXT_PRIMARY_KEY,
    'stack_info': TEXT
}

indexes = {
    "index1": ["step", "rank", "data_source", "up_node", "node_order"],
    "index2": ["step", "rank", "data_source", "node_name"],
    "index3": ["step", "rank", "data_source", "node_order"],
    "index4": ["step", "rank", "node_order"],
    "index5": ["step", "rank", "micro_step_id", "node_order"],
    "index6": ["step", "rank", "modified", "matched_node_link"],
    "index7": ["is_distributed"]
}

SAFE_NAME_PATTERN = re.compile(r'^[a-zA-Z0-9_]+$')


def is_safe_identifier(name):
    """验证标识符是否安全（防止SQL注入）"""
    return isinstance(name, str) and SAFE_NAME_PATTERN.match(name) is not None


def create_table_sql_from_dict(table_name, columns_dict):
    """
    根据提供的表名和列定义字典生成CREATE TABLE SQL语句。
    """
    if not is_safe_identifier(table_name):
        raise ValueError(f"Invalid table name: {table_name} - potential SQL injection risk!")

    sql = f"CREATE TABLE IF NOT EXISTS {table_name} (\n"

    column_definitions = []
    for column_name, column_type in columns_dict.items():
        if not is_safe_identifier(column_name):
            raise ValueError(f"Invalid column name: {column_name} - potential SQL injection risk!")

        column_definitions.append(f"    {column_name} {column_type}")

    sql += ",\n".join(column_definitions)
    sql += "\n);"

    return sql


def create_insert_sql_from_dict(table_name, columns_dict, ignore_insert=False):
    """
    根据提供的表名和数据字典生成INSERT INTO SQL语句。
    """
    if not is_safe_identifier(table_name):
        raise ValueError(f"Invalid table name: {table_name} - potential SQL injection risk!")

    columns = list(columns_dict.keys())

    for column_name in columns:
        if not is_safe_identifier(column_name):
            raise ValueError(f"Invalid column name: {column_name} - potential SQL injection risk!")

    placeholders = ["?"] * len(columns)

    columns_string = ", ".join(columns)
    placeholders_string = ", ".join(placeholders)

    sql_prefix = "INSERT OR IGNORE INTO" if ignore_insert else "INSERT INTO"
    sql = f"{sql_prefix} {table_name} ({columns_string}) VALUES ({placeholders_string})"
    return sql


def to_db(db_path, create_table_sql, insert_sql, data, pbar_info=None):
    max_retries = 10
    initial_delay = 0.1
    if not os.path.exists(db_path):
        check_path_before_create(db_path)
    else:
        FileChecker(db_path, FileCheckConst.FILE, FileCheckConst.READ_WRITE_ABLE,
                    FileCheckConst.DB_SUFFIX).common_check()

    retry_count = 0
    current_delay = initial_delay
    total_data = len(data)
    commited_data = 0

    if pbar_info:
        update_shared_dict(pbar_info.current_stage_dict, pbar_info.task_id, 1)

    while retry_count <= max_retries:
        conn = None
        try:
            conn = sqlite3.connect(db_path, timeout=30)
            cursor = conn.cursor()
            # 启用WAL模式提升多进程读写并发能力
            cursor.execute("PRAGMA journal_mode=WAL")
            cursor.execute("PRAGMA synchronous=NORMAL")
            cursor.execute(create_table_sql)
            for i in range(0, total_data, DB_INSERT_SIZE):
                batch = data[i:i + DB_INSERT_SIZE]
                cursor.executemany(insert_sql, batch)

                if pbar_info:
                    commited_data += len(batch)
                    update_pbar_info(pbar_info, commited_data, total_data)
            conn.commit()
            return
        except sqlite3.OperationalError as e:
            if "database is locked" in str(e).lower():
                retry_count += 1
                if retry_count > max_retries:
                    logger.error(f"Database lock conflict retry attempts exhausted ({max_retries}): {e}")
                    raise RuntimeError(f"DB lock retry exhausted: {e}") from e

                logger.warning(
                    f"DB lock conflict (retry {retry_count}/{max_retries}), wait {current_delay:.2f}s : {e}"
                )
                time.sleep(current_delay)
                current_delay *= 2
                continue

            logger.error(f"An sqlite3 error occurred: {e}")
            raise e
        except sqlite3.Error as e:
            logger.error(f"An sqlite3 error occurred: {e}")
            raise e
        except Exception as e:
            logger.error(f"An unknown error occurred: {e}")
            raise e
        finally:
            if conn:
                conn.close()


def add_table_index(db_path):
    FileChecker(db_path, FileCheckConst.FILE, FileCheckConst.READ_WRITE_ABLE, FileCheckConst.DB_SUFFIX).common_check()
    try:
        conn = sqlite3.connect(db_path)
    except sqlite3.Error as e:
        logger.error(f"Unable to create database connection: {e}")
        raise RuntimeError("Unable to create database connection") from e

    try:
        cursor = conn.cursor()
        for index_name, columns in indexes.items():
            if not is_safe_identifier(index_name):
                raise ValueError(f"Invalid index name: {index_name} - potential SQL injection risk!")

            for column in columns:
                if not is_safe_identifier(column):
                    raise ValueError(f"Invalid column name in index: {column} - potential SQL injection risk!")

            columns_str = ', '.join(columns)
            index_sql = f'''
                CREATE INDEX IF NOT EXISTS {index_name} ON tb_nodes ({columns_str});
            '''
            cursor.execute(index_sql)
        conn.commit()
    except sqlite3.Error as e:
        logger.error(f"Failed to add table index: {e}")
        raise RuntimeError("Failed to add table index") from e
    finally:
        conn.close()


def post_process_db(db_path, pbar_info=None, is_parallel_merge=False):
    logger.info('Start adding index to db file, please wait...')
    add_table_index(db_path)
    change_mode(db_path, FileCheckConst.DATA_FILE_AUTHORITY)
    if pbar_info:
        post_process_db_pbar(pbar_info, is_parallel_merge)
    logger.info('Adding index to db file completed.')


def node_to_db(graph, db_name, pbar_info=None):
    create_table_sql = create_table_sql_from_dict('tb_nodes', node_columns)
    insert_sql = create_insert_sql_from_dict('tb_nodes', node_columns)
    data = []
    stack_dict = {}
    for i, node in enumerate(graph.get_sorted_nodes()):
        stack_info_text = json.dumps(node.stack_info)
        if stack_info_text not in stack_dict:
            stack_dict[stack_info_text] = get_stack_unique_id(graph, stack_dict)
        data.append((get_node_unique_id(graph, node), get_graph_unique_id(graph), i, node.id, node.op.value,
                     node.upnode.id if node.upnode else '',
                     json.dumps([node.id for node in node.subnodes]) if node.subnodes else '',
                     node.data.get(GraphConst.JSON_INDEX_KEY), node.data.get(GraphConst.OVERFLOW_LEVEL),
                     node.micro_step_id if node.micro_step_id is not None else 0, json.dumps(node.matched_node_link),
                     stack_dict.get(stack_info_text),
                     json.dumps(node.parallel_merge_info) if node.parallel_merge_info else '',
                     json.dumps(node.matched_distributed), 0,
                     json.dumps(format_node_data(node.input_data, node.id, graph.compare_mode)),
                     json.dumps(format_node_data(node.output_data, node.id, graph.compare_mode)),
                     graph.data_source, graph.data_path, graph.step, graph.rank,
                     1 if node.id.startswith(Const.DISTRIBUTED) else 0))
    to_db(db_name, create_table_sql, insert_sql, data, pbar_info=pbar_info)
    stack_to_db(stack_dict, db_name)


def config_to_db(config, db_name):
    create_table_sql = create_table_sql_from_dict('tb_config', config_columns)
    insert_sql = create_insert_sql_from_dict('tb_config', config_columns, ignore_insert=True)
    data = [("1", "compare" if config.graph_b else "build", config.task, config.tool_tip, config.micro_steps,
             config.overflow_check, json.dumps(config.node_colors), json.dumps(config.rank_list),
             json.dumps(config.step_list))]
    to_db(db_name, create_table_sql, insert_sql, data)


def stack_to_db(stack_dict, db_name):
    create_table_sql = create_table_sql_from_dict('tb_stack', stack_columns)
    insert_sql = create_insert_sql_from_dict('tb_stack', stack_columns)
    data = []
    for stack_info_text, unique_id in stack_dict.items():
        data.append((unique_id, stack_info_text))
    to_db(db_name, create_table_sql, insert_sql, data)


def get_graph_unique_id(graph):
    return f'{graph.data_source}_{graph.step}_{graph.rank}'


def get_node_unique_id(graph, node):
    return f'{get_graph_unique_id(graph)}_{node.id}'


def get_stack_unique_id(graph, stack_dict):
    return f'{get_graph_unique_id(graph)}_{len(stack_dict)}'


class DBActuator:
    def __init__(self, db_path):
        self.db_path = db_path
        self._init_connection()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            try:
                self.conn.close()
                logger.info(f"Database connection to {self.db_path} closed successfully")
            except sqlite3.Error as e:
                logger.error(f"Failed to close database connection: {e}")
        return False

    def query_distributed_nodes_info(self):
        query = """
 	         SELECT
 	             node_name,
 	             input_data,
 	             rank,
 	             step,
 	             data_source,
 	             precision_index,
 	             overflow_level
 	         FROM
 	             tb_nodes
 	         WHERE
 	             is_distributed = 1
 	         """
        if not self.conn:
            logger.warning("Database connection is not initialized.")
            return []
        try:
            with self.conn as conn:
                cursor = conn.execute(query)
                rows = cursor.fetchall()
                cursor.close()

            return rows
        except Exception as e:
            logger.error(f'Failed to query distributed nodes: {e}')
            return []

    def update_matched_distributed(self, distributed_info: Dict[int, dict], step=0,
                                   data_source=GraphConst.JSON_NPU_KEY, batch_size=1000):
        update_data_list = []
        update_sql = """
     	         UPDATE tb_nodes 
     	         SET matched_distributed = ? 
     	         WHERE step = ? AND rank = ? AND data_source = ? AND node_name = ?;

     	         """
        if not self.conn:
            logger.warning("Database connection is not initialized.")
            return
        for rank, nodes_dict in distributed_info.items():
            for node in nodes_dict.values():
                if not node.matched_distributed:
                    continue
                try:

                    matched_distributed = json.dumps(node.matched_distributed)
                except Exception as e:
                    logger.warning(f"Failed to serialize matched_distributed for node {node.id}: {e}")
                    continue
                update_data_list.append((matched_distributed, step, rank, data_source, node.id))

        try:
            with self.conn as conn:
                cursor = conn.cursor()
                for i in range(0, len(update_data_list), batch_size):
                    batch_data = update_data_list[i:i + batch_size]
                    cursor.executemany(update_sql, batch_data)
                cursor.close()

        except Exception as e:
            logger.error(f"Failed to update matched_distributed: {e}")
            raise

    def _init_connection(self):
        FileChecker(self.db_path, FileCheckConst.FILE, FileCheckConst.READ_WRITE_ABLE,
                    FileCheckConst.DB_SUFFIX).common_check()
        try:
            self.conn = sqlite3.connect(self.db_path)
        except sqlite3.Error as e:
            logger.error(f"Failed to connect to database: {e}")
            raise e



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch

from msprobe.core.common.const import Const
from msprobe.pytorch.common.log import logger

torch_version_above_or_equal_2 = torch.__version__.split('+')[0] >= '2.0'
if torch_version_above_or_equal_2:
    from torch.optim.optimizer import register_optimizer_step_pre_hook, register_optimizer_step_post_hook


def register_optimizer_hook(data_collector):
    def optimizer_pre_step_hook(optimizer, args, kwargs):
        data_collector.optimizer_status = Const.OPTIMIZER

    def optimizer_post_step_hook(optimizer, args, kwargs):
        data_collector.optimizer_status = Const.END_PREFIX + Const.OPTIMIZER

    def patch_clip_grad(func):
        def wrapper(*args, **kwargs):
            data_collector.optimizer_status = Const.CLIP_GRAD
            result = func(*args, **kwargs)
            data_collector.optimizer_status = Const.END_PREFIX + Const.CLIP_GRAD
            return result

        return wrapper

    if torch_version_above_or_equal_2:
        register_optimizer_step_pre_hook(optimizer_pre_step_hook)
        register_optimizer_step_post_hook(optimizer_post_step_hook)
    else:
        logger.info_on_rank_0("Pytorch version is below 2.0, cannot register optimizer hook.")

    try:
        torch.nn.utils.clip_grad_norm_ = patch_clip_grad(torch.nn.utils.clip_grad_norm_)
        torch.nn.utils.clip_grad_norm = patch_clip_grad(torch.nn.utils.clip_grad_norm)
        torch.nn.utils.clip_grad_value_ = patch_clip_grad(torch.nn.utils.clip_grad_value_)
    except Exception as e:
        logger.info_on_rank_0("Cannot patch clip grad function. detail:%s" % str(e))

    try:
        from megatron.core.optimizer import MegatronOptimizer
        MegatronOptimizer.clip_grad_norm = patch_clip_grad(MegatronOptimizer.clip_grad_norm)
    except ImportError:
        pass
    except Exception as e:
        logger.info_on_rank_0("Cannot patch megatron clip grad function. detail:%s" % str(e))


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import itertools
import os
from collections import defaultdict
from dataclasses import dataclass

import pandas as pd
import torch
from torch.utils.tensorboard import SummaryWriter

from msprobe.core.common.const import FileCheckConst, MonitorConst
from msprobe.core.common.file_utils import change_mode, create_directory, write_df_to_csv
from msprobe.core.monitor.anomaly_processor import AnomalyDataFactory, AnomalyTurbulence, AnomalyScanner
from msprobe.pytorch.common.log import logger


class BCOLORS:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


@dataclass
class WriterInput:
    path: str
    ad_rules: list
    job_id: str
    anomaly_factory: AnomalyDataFactory = None
    ndigits: int = 6
    step_count_per_record: int = 1


class BaseWriterWithAD:
    def __init__(self, writer_input: WriterInput):
        self.tag2scalars = {}
        self.ad_rules = writer_input.ad_rules
        self.job_id = writer_input.job_id
        self.anomaly_factory = writer_input.anomaly_factory
        self.anomalies = []
        self.ndigits = writer_input.ndigits
        self.beta = 0.99

    @staticmethod
    def stack_tensors(tensor_list):
        """
        Torch not support stack cpu and xpu tensors. Group the tensors into cpu_group and xpu_group,
        stack them separately, migrate xpu_group to cpu, and then restore in the order of input.

        :param tensor_list: [tensor(-1.6165), tensor(-1.0985), tensor(-1.7777), tensor(-1.8408, device='npu:0')]
        :return: result: list of float
        """
        cpu_tensors = []
        xpu_tensors = []

        for tensor in tensor_list:
            if isinstance(tensor, torch.Tensor) and tensor.device.type != 'cpu':
                # 将device上的tensor先stack后to cpu
                xpu_tensors.append(tensor)
            else:
                cpu_tensors.append(tensor)

        xpu_stack = torch.stack(xpu_tensors).cpu() if xpu_tensors else torch.tensor([])

        # 按照输入的顺序恢复
        result = []
        cpu_tensors_idx, xpu_tensors_idx = 0, 0
        for tensor in tensor_list:
            if isinstance(tensor, torch.Tensor) and tensor.device.type != 'cpu':
                result.append(xpu_stack[xpu_tensors_idx])
                xpu_tensors_idx += 1
            else:
                result.append(cpu_tensors[cpu_tensors_idx])
                cpu_tensors_idx += 1

        return result

    def get_anomalies(self):
        """返回已检测到的异常列表
        """
        return self.anomalies

    def clear_anomalies(self):
        self.anomalies.clear()

    def add_scalar(self, tag, scalar_value, global_step=None):
        """If an anomaly is detected, the anomaly information is recorded and added to self.anomalies.
        Args:
            tag (tuple): tuple of tag_name and tag like ('0:1.post_attention_norm.weight/rank0/pre_grad', 'min').
            scalar_value (float): scalar_value.
            global_step (int): global_step.
        Returns:
            None
        """
        if not self.ad_rules or tag[-1] in ["shape", "dtype"]:
            return
        if isinstance(scalar_value, torch.Tensor):
            scalar_value = scalar_value.item()
        avg = self._update_tag2scalars(tag, scalar_value)
        detected, rule_name = self._ad(scalar_value, history=avg)
        if detected:
            if rule_name == AnomalyTurbulence.name and tag[-1] not in ["norm", "mean"]:
                return
            exception_message = (f"Rule {rule_name} reports anomaly signal in {tag} at step {global_step}, "
                                 f"current value {scalar_value}, history mean {avg}.")
            logger.info(f"{BCOLORS.WARNING}> {exception_message}{BCOLORS.ENDC}")
            # append to self.anomalies for dump
            if self.anomaly_factory:
                self.anomalies.append(self.anomaly_factory.create(tag, exception_message, global_step))

    def write_metrics(self, ops, metric_value, step, prefix=''):
        if not metric_value:
            return
        tensors = []
        tags = list(itertools.product(metric_value.keys(), ops))
        for op2tensor in metric_value.values():
            tensors.extend(op2tensor.values())
        if not tensors:
            return

        n_slices = len(tensors) // MonitorConst.SLICE_SIZE
        with torch.no_grad():
            for i in range(n_slices + 1):
                begin = i * MonitorConst.SLICE_SIZE
                end = (i + 1) * MonitorConst.SLICE_SIZE
                if begin == len(tensors):
                    continue
                metric_list = self.stack_tensors(tensors[begin:end])
                for tag, metric in zip(tags[begin:end], metric_list):
                    self.add_scalar(tag, metric, step)

    def _ad(self, scalar_value, history):
        return AnomalyScanner.scan(self.ad_rules, history, cur=scalar_value)

    def _update_tag2scalars(self, tag, scalar_value):
        """Update the average and count of a scalar value associated with a tag.

        This method is used to maintain a running average of scalar values for each tag.


        Args:
            tag (str): The tag identifier.
            scalar_value (float): The scalar value to be added.

        Returns:
            float: The average value before update.
        """
        abs_scalar_value = abs(scalar_value)
        if tag not in self.tag2scalars:
            self.tag2scalars[tag] = {'avg': abs_scalar_value, 'count': 0}
        avg = self.tag2scalars[tag]['avg']
        self.tag2scalars[tag]['avg'] = self.beta * avg + (1 - self.beta) * abs_scalar_value
        self.tag2scalars[tag]['count'] += 1
        return avg


class CSVWriterWithAD(BaseWriterWithAD):
    def __init__(self, writer_input: WriterInput):
        super().__init__(writer_input)

        path = writer_input.path
        self.log_dir = path
        create_directory(path)
        change_mode(path, FileCheckConst.DATA_DIR_AUTHORITY)
        self.context_dict = defaultdict(list)
        self.header = []
        self.step_count_per_record = writer_input.step_count_per_record

    def get_step_interval(self, step):
        count = step // self.step_count_per_record
        return count * self.step_count_per_record, (count + 1) * self.step_count_per_record - 1

    def write_csv(self, prefix, step):
        """
        Args:
            prefix[str]: prefix of output csv file e.g. grad_unreduced
            step[int]
        """
        if len(self.context_dict) == 0:
            return

        ster_start, step_end = self.get_step_interval(step)
        filepath = os.path.join(self.log_dir, f'{prefix}_{ster_start}-{step_end}.csv')
        if not os.path.exists(filepath):
            data_frame = pd.DataFrame(columns=self.header)
            write_df_to_csv(data_frame, filepath)

        new_data = []
        for name, metric_value in self.context_dict.items():
            new_line = name.split(MonitorConst.NAME_SEP) + metric_value
            new_line.insert(2, step)
            new_data.append(new_line)
        new_data = pd.DataFrame(new_data).round(self.ndigits).fillna("nan")
        write_df_to_csv(new_data, filepath, mode='a+', header=False)
        self.context_dict = defaultdict(list)

    def add_scalar(self, tag, scalar_value, global_step):
        """
        ('0:1.post_attention_norm.weight/rank0/pre_grad', 'min')
        """
        super().add_scalar(tag, scalar_value, global_step)

        name = tag[0].split('/')[0]
        if isinstance(scalar_value, torch.Tensor):
            value = scalar_value.item()
        elif isinstance(scalar_value, torch.Size):
            value = list(scalar_value)
        else:
            value = scalar_value
        self.context_dict[name].append(value)

    def write_metrics(self, ops, metric_value, step, prefix='', **kwargs):
        super().write_metrics(ops, metric_value, step, prefix='')

        if prefix in [MonitorConst.ACTV, MonitorConst.ACTVGRAD] or kwargs.get("use_micro_step"):
            self.header = MonitorConst.CSV_HEADER_MICRO_STEP + ops
        else:
            self.header = MonitorConst.CSV_HEADER + ops
        self.write_csv(prefix, step)

    def close(self):
        pass


class SummaryWriterWithAD(SummaryWriter, BaseWriterWithAD):
    def __init__(self, writer_input: WriterInput):

        path = writer_input.path
        if not os.path.exists(path):
            create_directory(path)
        try:
            super(SummaryWriter, self).__init__(writer_input)
            super().__init__(path)
        except Exception as e:
            logger.error(f'error when init summary writer at {path}: {e}')
            raise ValueError("Init summary writer error.") from e

    def add_scalar(self, tag, scalar_value, global_step):
        super(SummaryWriter, self).add_scalar(tag, scalar_value, global_step)
        tag = f'{tag[0]}_{tag[1]}'
        super().add_scalar(tag, scalar_value, global_step)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch
import numpy as np
import matplotlib.pyplot as plt
from msprobe.pytorch.monitor.features import cal_histc


class HeatmapVisualizer:
    def __init__(self) -> None:
        self.histogram_bins_num = 30
        self.min_val = -1
        self.max_val = 1
        self.histogram_edges = None
        self.histogram_sum_data_np = None  # matrix shape is [bins_num * total_step]
        self.cur_step_histogram_data = None
        self.histogram_edges = torch.linspace(self.min_val, self.max_val, self.histogram_bins_num)

    def pre_cal(self, tensor):
        self.cur_step_histogram_data = cal_histc(tensor_cal=tensor, bins_total=self.histogram_bins_num,
                                                 min_val=self.min_val, max_val=self.max_val)

    def visualize(self, tag_name: str, step, summary_writer):
        if self.histogram_sum_data_np is None or self.histogram_sum_data_np.size == 0:
            self.histogram_sum_data_np = np.expand_dims(self.cur_step_histogram_data.cpu(), 0).T
        else:
            # add new data along a different axis because we transposed early
            # matrix shape is [bins_num * total_step]
            self.histogram_sum_data_np = np.concatenate((self.histogram_sum_data_np, np.expand_dims(
                self.cur_step_histogram_data.cpu(), 1)), axis=1)

        fig, ax = plt.subplots()
        cax = ax.matshow(self.histogram_sum_data_np, cmap='hot', aspect='auto')
        fig.colorbar(cax)

        lbs = [f'{self.histogram_edges[i]:.2f}' for i in range(self.histogram_bins_num)]
        plt.yticks(ticks=range(self.histogram_bins_num), labels=lbs)
        ax.set_xlabel('Step')
        ax.set_ylabel('Value Range')
        plt.title(f'Total Step: {step}')

        # Convert matplotlib figure to an image format suitable for TensorBoard
        fig.canvas.draw()
        image = torch.from_numpy(np.array(fig.canvas.renderer.buffer_rgba()))
        plt.close(fig)
        summary_writer.add_image(tag_name, image.permute(2, 0, 1), global_step=step, dataformats='CHW')



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import inspect
from typing import Dict, Any, Optional, Callable, Union, List, Tuple

from msprobe.core.common.const import Const
from msprobe.core.common.file_utils import load_yaml
from msprobe.core.common.log import logger


def _get_attr(module, attr_name):
    if Const.SEP in attr_name:
        sub_module_name, sub_attr = attr_name.rsplit(Const.SEP, 1)
        sub_module = getattr(module, sub_module_name, None)
        attr = getattr(sub_module, sub_attr, None)
    else:
        attr = getattr(module, attr_name, None)
    return attr


class ApiWrapper:
    def __init__(
        self, api_types: Dict[str, Dict[str, Any]],
        api_list_paths: Union[str, List[str], Tuple[str]],
        blacklist: Union[List[str], Tuple[str]] = None
    ):
        self.api_types = api_types
        if not isinstance(api_list_paths, (list, tuple)):
            api_list_paths = [api_list_paths] * len(self.api_types)
        elif len(api_list_paths) != len(self.api_types):
            raise RuntimeError("The number of api_list_paths must be equal to the number of frameworks in 'api_types', "
                               "when api_list_paths is a list or tuple.")
        self.api_list_paths = api_list_paths
        self.blacklist = blacklist if blacklist else []
        self.api_names = self._get_api_names()
        self.wrapped_api_functions = dict()

    @staticmethod
    def deal_with_self_kwargs(api_name, api_func, args, kwargs):
        if kwargs and 'self' in kwargs:
            func_params = None
            try:
                func_params = inspect.signature(api_func).parameters
            except Exception:
                if api_name in Const.API_WITH_SELF_ARG:
                    func_params = inspect.signature(Const.API_WITH_SELF_ARG.get(api_name)).parameters
            if func_params is None:
                return False, args, kwargs

            for name, param in func_params.items():
                if name == 'self' and param.kind == inspect.Parameter.KEYWORD_ONLY:
                    return False, args, kwargs
            args_ = list(args)
            names_and_values = []
            self_index = 0
            for i, item in enumerate(func_params.items()):
                names_and_values.append((item[0], item[1].default))
                if item[0] == 'self':
                    self_index = i
                    break
            for i in range(len(args), self_index + 1):
                if names_and_values[i][0] in kwargs:
                    args_.append(kwargs.pop(names_and_values[i][0]))
                else:
                    args_.append(names_and_values[i][1])
            args = tuple(args_)

        return True, args, kwargs

    def wrap_api_func(self, api_name, api_func, prefix, hook_build_func, api_template):
        api_instance = api_template(api_name, api_func, prefix, hook_build_func)

        def api_function(*args, **kwargs):
            api_name_with_prefix = prefix + Const.SEP + str(api_name.split(Const.SEP)[-1])
            enable_wrap, args, kwargs = self.deal_with_self_kwargs(api_name_with_prefix, api_func, args, kwargs)
            if not enable_wrap:
                logger.warning(f'Cannot collect precision data of {api_name_with_prefix}. '
                               'It may be fixed by passing the value of "self" '
                               'as a positional argument instead of a keyword argument. ')
                return api_func(*args, **kwargs)
            return api_instance(*args, **kwargs)

        for attr_name in Const.API_ATTR_LIST:
            if hasattr(api_func, attr_name):
                attr_value = getattr(api_func, attr_name)
                setattr(api_function, attr_name, attr_value)

        return api_function

    def wrap_api(
        self, api_templates, hook_build_func: Optional[Callable]
    ):
        api_types_num = sum([len(v) for v in self.api_types.values()])
        if not isinstance(api_templates, (list, tuple)):
            api_templates = [api_templates] * api_types_num
        elif len(api_templates) != api_types_num:
            raise RuntimeError("The number of api_templates must be equal to the number of api_types, "
                               "when api_templates is a list or tuple.")

        self.wrapped_api_functions.clear()
        index = 0
        for framework, api_types in self.api_types.items():
            wrapped_functions_in_framework = dict()
            for api_type, api_modules in api_types.items():
                wrapped_functions = dict()
                name_prefix = Const.API_DATA_PREFIX.get(framework, {}).get(api_type, "API")
                api_template = api_templates[index]
                index += 1
                for api_name in self.api_names.get(framework, {}).get(api_type, []):
                    ori_api = None
                    for module in api_modules[0]:
                        ori_api = ori_api or _get_attr(module, api_name)
                    if callable(ori_api):
                        wrapped_functions[api_name] = self.wrap_api_func(
                            api_name,
                            ori_api,
                            name_prefix,
                            hook_build_func,
                            api_template
                        )
                wrapped_functions_in_framework[api_type] = wrapped_functions
            self.wrapped_api_functions[framework] = wrapped_functions_in_framework
        return self.wrapped_api_functions

    def _get_api_names(self):
        api_names = dict()

        for index, framework in enumerate(self.api_types.keys()):
            api_list = load_yaml(self.api_list_paths[index])
            valid_names = dict()
            for api_type, api_modules in self.api_types.get(framework, {}).items():
                key_in_file = Const.SUPPORT_API_DICT_KEY_MAP.get(framework, {}).get(api_type)
                api_from_file = api_list.get(key_in_file, [])
                names = set()
                for api_name in api_from_file:
                    if f'{key_in_file}.{api_name}' in self.blacklist:
                        continue
                    target_attr = api_name
                    for module in api_modules[0]:
                        if Const.SEP in api_name:
                            sub_module_name, target_attr = api_name.rsplit(Const.SEP, 1)
                            target_module = getattr(module, sub_module_name, None)
                        else:
                            target_module = module
                        if target_module and target_attr in dir(target_module):
                            names.add(api_name)
                valid_names[api_type] = names
            api_names[framework] = valid_names

        return api_names


class ApiRegistry:
    """
    Base class for api registry.
    """

    def __init__(self, api_types, inner_used_api, supported_api_list_path, api_templates, blacklist=None):
        self.ori_api_attr = dict()
        self.wrapped_api_attr = dict()
        self.inner_used_ori_attr = dict()
        self.inner_used_wrapped_attr = dict()
        self.api_types = api_types
        self.inner_used_api = inner_used_api
        self.supported_api_list_path = supported_api_list_path
        self.api_templates = api_templates
        self.blacklist = blacklist if blacklist else []
        self.all_api_registered = False

    @staticmethod
    def store_ori_attr(ori_api_groups, api_list, api_ori_attr):
        for api in api_list:
            ori_api = None
            for ori_api_group in ori_api_groups:
                ori_api = ori_api or _get_attr(ori_api_group, api)
            api_ori_attr[api] = ori_api

    @staticmethod
    def set_api_attr(api_group, attr_dict):
        for api, api_attr in attr_dict.items():
            if Const.SEP in api:
                sub_module_name, sub_op = api.rsplit(Const.SEP, 1)
                sub_module = getattr(api_group, sub_module_name, None)
                if sub_module is not None:
                    setattr(sub_module, sub_op, api_attr)
            else:
                setattr(api_group, api, api_attr)

    @staticmethod
    def register_custom_api(module, api_name, api_prefix, hook_build_func, api_template):
        def wrap_api_func(api_name, api_func, prefix, hook_build_func, api_template):
            def api_function(*args, **kwargs):
                return api_template(api_name, api_func, prefix, hook_build_func)(*args, **kwargs)

            api_function.__name__ = api_name
            return api_function

        setattr(module, api_name,
                wrap_api_func(api_name, getattr(module, api_name), api_prefix, hook_build_func, api_template))

    def register_all_api(self):
        self.all_api_registered = True
        for framework, api_types in self.api_types.items():
            for api_type, api_modules in api_types.items():
                api_type_with_framework = framework + Const.SEP + api_type
                for module in api_modules[1]:
                    self.set_api_attr(module, self.wrapped_api_attr.get(api_type_with_framework, {}))

    def register_inner_used_api(self):
        for api_type in self.inner_used_api.keys():
            self.set_api_attr(self.inner_used_api.get(api_type)[0], self.inner_used_wrapped_attr.get(api_type, {}))

    def restore_all_api(self):
        self.all_api_registered = False
        for framework, api_types in self.api_types.items():
            for api_type, api_modules in api_types.items():
                api_type_with_framework = framework + Const.SEP + api_type
                for module in api_modules[1]:
                    self.set_api_attr(module, self.ori_api_attr.get(api_type_with_framework, {}))

    def restore_inner_used_api(self):
        for api_type in self.inner_used_api.keys():
            self.set_api_attr(self.inner_used_api.get(api_type)[0], self.inner_used_ori_attr.get(api_type, {}))

    def initialize_hook(self, hook_build_func):
        api_wrapper = ApiWrapper(self.api_types, self.supported_api_list_path, self.blacklist)
        wrapped_api_functions = api_wrapper.wrap_api(self.api_templates, hook_build_func)

        for framework, api_types in self.api_types.items():
            for api_type, api_modules in api_types.items():
                ori_attr = dict()
                self.store_ori_attr(api_modules[0], api_wrapper.api_names.get(framework).get(api_type), ori_attr)
                api_type_with_framework = framework + Const.SEP + api_type
                self.ori_api_attr[api_type_with_framework] = ori_attr
                self.wrapped_api_attr[api_type_with_framework] = wrapped_api_functions.get(framework).get(api_type)

        for inner_used_api_type, inner_used_api_list in self.inner_used_api.items():
            ori_attr = dict()
            wrapped_attr = dict()
            for api_name in inner_used_api_list[1:]:
                if self.ori_api_attr.get(inner_used_api_type, {}).get(api_name):
                    ori_attr[api_name] = self.ori_api_attr.get(inner_used_api_type).get(api_name)
                    wrapped_attr[api_name] = self.wrapped_api_attr.get(inner_used_api_type).get(api_name)
            self.inner_used_ori_attr[inner_used_api_type] = ori_attr
            self.inner_used_wrapped_attr[inner_used_api_type] = wrapped_attr



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


from abc import ABC, abstractmethod
from msprobe.core.compare.indicator_analysis.api_data import ApiData
from msprobe.core.compare.indicator_analysis.utils import is_inf_or_nan, str2float, ResultLevel, IgnoreInfo, \
    get_data_list_by_ignore_info
from msprobe.core.common.const import CompareConst


class BaseAlgorithm(ABC):
    """比对算法基类"""

    @abstractmethod
    def run(self, api_data: ApiData, ignore_info: IgnoreInfo):
        """
        算法执行接口
        :param api_data: 结构化的 API 数据
        :param ignore_info: 当前 API 数据需要忽略的指标信息
        """
        pass


class InfNanErrChecker(BaseAlgorithm):
    """
    适用于真实数据模式、统计数据模式
    一个 API 或模块的 NPU 的最大值或最小值中存在 nan/inf/-inf 标记为 error
    但如果 bench 侧也有相同现象，则忽略
    """

    def __init__(self):
        self.result_level = ResultLevel.ERROR
        self.err_msg = f'{self.result_level.value}: There is nan/inf/-inf in the maximum or minimum value of NPU.'

    def run(self, api_data: ApiData, ignore_info: IgnoreInfo):
        data_lists = get_data_list_by_ignore_info(api_data, ignore_info)
        if not data_lists:
            return

        for data_list in data_lists:
            bench_max = api_data.get_data_by_header(CompareConst.BENCH_MAX, data_list)
            bench_min = api_data.get_data_by_header(CompareConst.BENCH_MIN, data_list)

            if is_inf_or_nan(bench_max) or is_inf_or_nan(bench_min):
                continue

            npu_max = api_data.get_data_by_header(CompareConst.NPU_MAX, data_list)
            npu_min = api_data.get_data_by_header(CompareConst.NPU_MIN, data_list)

            if is_inf_or_nan(npu_max) or is_inf_or_nan(npu_min):
                api_data.set_result(data_list, self.result_level)
                api_data.set_err_msg(data_list, self.err_msg)


class RelativeErrChecker(BaseAlgorithm):
    """
    适用于统计数据模式
    指标需要结合输入和输出共同计算得到
    一个 API 或模块的 input 的相对误差 < 0.1 且 output 的相对误差 > 0.5，默认选取norm relative err观测， 标记为 error
    """

    def __init__(self):
        self.in_threshold = 0.1
        self.out_threshold = 0.5
        self.result_level = ResultLevel.ERROR
        self.err_msg = (f'{self.result_level.value}: The {CompareConst.NORM_RELATIVE_ERR} of output '
                        f'is greater than {self.out_threshold}.')

    def run(self, api_data: ApiData, ignore_info: IgnoreInfo):
        if ignore_info in [IgnoreInfo.ALL_IGNORE, IgnoreInfo.INPUT_IGNORE]:
            return

        norm_relative_err_max = abs(api_data.get_min_or_max_value(CompareConst.NORM_RELATIVE_ERR, is_min=False))

        if norm_relative_err_max < self.in_threshold:
            for data_list in api_data.output_data:
                norm_relative_err = str2float(api_data.get_data_by_header(CompareConst.NORM_RELATIVE_ERR, data_list))
                if abs(norm_relative_err) > self.out_threshold:
                    api_data.set_result(data_list, self.result_level)
                    api_data.set_err_msg(data_list, self.err_msg)


class OneThousandthErrChecker(BaseAlgorithm):
    """
    适用于真实数据模式
    指标需要结合输入和输出共同计算得到
    一个 API 或模块的 One Thousandth Err Ratio 的 input/parameters > 0.9 同时 output < 0.6， 标记为 error
    """

    def __init__(self):
        self.input_threshold = 0.9
        self.output_threshold = 0.1
        self.result_level = ResultLevel.ERROR
        self.err_msg = (f'{self.result_level.value}: The input/parameters of '
                        f'One Thousandth Err Ratio > 0.9 while the output < 0.6.')

    def run(self, api_data: ApiData, ignore_info: IgnoreInfo):
        if ignore_info in [IgnoreInfo.ALL_IGNORE, IgnoreInfo.INPUT_IGNORE]:
            return
        if not api_data.output_data:
            return

        min_input_ratio = api_data.get_min_or_max_value(CompareConst.ONE_THOUSANDTH_ERR_RATIO)
        min_output_ratio = api_data.get_min_or_max_value(CompareConst.ONE_THOUSANDTH_ERR_RATIO, is_input=False)

        if min_input_ratio > self.input_threshold and min_output_ratio < self.output_threshold:
            api_data.set_result(api_data.output_data[0], self.result_level)
            api_data.set_err_msg(api_data.output_data[0], self.err_msg)


class RequiresGradErrChecker(BaseAlgorithm):
    """
    适用于真实数据模式、统计数据模式
    一个 API 或模块的 Requires_grad Consistent 为 False，标记为 error
    """

    def __init__(self):
        self.result_level = ResultLevel.ERROR
        self.err_msg = f'{self.result_level.value}: The Required_Grad of NPU and Bench are inconsistent'

    def run(self, api_data: ApiData, ignore_info: IgnoreInfo):
        data_lists = get_data_list_by_ignore_info(api_data, ignore_info)
        if not data_lists:
            return

        for data_list in data_lists:
            # not match
            bench_name = api_data.get_data_by_header(CompareConst.BENCH_NAME, data_list)
            if not bench_name or bench_name == CompareConst.N_A:
                continue

            npu_req_grad = api_data.get_data_by_header(CompareConst.NPU_REQ_GRAD, data_list)
            bench_req_grad = api_data.get_data_by_header(CompareConst.BENCH_REQ_GRAD, data_list)

            if not npu_req_grad or not bench_req_grad:
                continue

            if npu_req_grad != bench_req_grad:
                api_data.set_result(data_list, self.result_level)
                api_data.set_err_msg(data_list, self.err_msg)


class ParametersErrChecker(BaseAlgorithm):
    """
    适用于真实数据模式、统计数据模式
    一个 API 或模块的非 tensor 标量参数，NPU 和 Bench 不一致，标记为 error
    """

    def __init__(self):
        self.result_level = ResultLevel.ERROR
        self.err_msg = f'{self.result_level.value}: The scalar parameters of NPU and Bench are inconsistent.'

    def run(self, api_data: ApiData, ignore_info: IgnoreInfo):
        if ignore_info in [IgnoreInfo.ALL_IGNORE, IgnoreInfo.INPUT_IGNORE]:
            return

        for data_list in api_data.input_data:
            # not match
            bench_name = api_data.get_data_by_header(CompareConst.BENCH_NAME, data_list)
            if not bench_name or bench_name == CompareConst.N_A:
                continue

            npu_dtype = api_data.get_data_by_header(CompareConst.NPU_DTYPE, data_list)
            bench_dtype = api_data.get_data_by_header(CompareConst.BENCH_DTYPE, data_list)
            if not npu_dtype or not bench_dtype:
                continue
            # 非tensor标量的dtype一定包含'class'，例如int类型的dtype为<class int>
            if 'class' not in npu_dtype or 'class' not in bench_dtype:
                continue

            npu_shape = api_data.get_data_by_header(CompareConst.NPU_SHAPE, data_list)
            # 以shape是否为[]判断其是否为标量
            if str(npu_shape) != '[]':
                continue
            npu_max = api_data.get_data_by_header(CompareConst.NPU_MAX, data_list)
            bench_max = api_data.get_data_by_header(CompareConst.BENCH_MAX, data_list)
            if npu_max != bench_max:
                api_data.set_result(data_list, self.result_level)
                api_data.set_err_msg(data_list, self.err_msg)


class CRC32ErrChecker(BaseAlgorithm):
    """
    适用于MD5模式
    NPU 与标杆的 CRC-32 值不一致，标记为 error
    NPU 与标杆的参数未匹配上，标记为 warning
    """

    def __init__(self):
        self.err_level = ResultLevel.ERROR
        self.err_msg = f'{self.err_level.value}: The CRC-32 value of NPU differs from that of the bench.'
        self.warn_level = ResultLevel.WARNING
        self.warn_msg = f'{self.warn_level.value}: The parameter of NPU does not match the bench.'

    def run(self, api_data: ApiData, ignore_info: IgnoreInfo):
        data_lists = get_data_list_by_ignore_info(api_data, ignore_info)
        if not data_lists:
            return

        null_set = (CompareConst.N_A, CompareConst.NAN)
        for data_list in data_lists:
            npu_md5 = api_data.get_data_by_header(CompareConst.NPU_MD5, data_list)
            bench_md5 = api_data.get_data_by_header(CompareConst.BENCH_MD5, data_list)
            if npu_md5 != bench_md5:
                # 参数未匹配上
                if npu_md5 in null_set or bench_md5 in null_set:
                    api_data.set_result(data_list, self.warn_level)
                    api_data.set_err_msg(data_list, self.warn_msg)
                else:
                    api_data.set_result(data_list, self.err_level)
                    api_data.set_err_msg(data_list, self.err_msg)


class DTypeErrChecker(BaseAlgorithm):
    """
    适用于真实数据模式、统计数据模式
    一个 API 或模块的 dtype 不一致，标记为 error
    """

    def __init__(self):
        self.result_level = ResultLevel.ERROR
        self.err_msg = f'{self.result_level.value}: The dtype of NPU and Bench are inconsistent.'

    def run(self, api_data: ApiData, ignore_info: IgnoreInfo):
        data_lists = get_data_list_by_ignore_info(api_data, ignore_info)
        if not data_lists:
            return

        for data_list in data_lists:
            # not match
            bench_name = api_data.get_data_by_header(CompareConst.BENCH_NAME, data_list)
            if not bench_name or bench_name == CompareConst.N_A:
                continue

            npu_dtype = api_data.get_data_by_header(CompareConst.NPU_DTYPE, data_list)
            bench_dtype = api_data.get_data_by_header(CompareConst.BENCH_DTYPE, data_list)

            if not npu_dtype or not bench_dtype:
                continue

            if npu_dtype != bench_dtype:
                api_data.set_result(data_list, self.result_level)
                api_data.set_err_msg(data_list, self.err_msg)


class ShapeErrChecker(BaseAlgorithm):
    """
    适用于真实数据模式、统计数据模式
    一个 API 或模块的 shape 不一致，标记为 error
    """

    def __init__(self):
        self.result_level = ResultLevel.ERROR
        self.err_msg = f'{self.result_level.value}: The shape of NPU and Bench are inconsistent.'

    def run(self, api_data: ApiData, ignore_info: IgnoreInfo):
        data_lists = get_data_list_by_ignore_info(api_data, ignore_info)
        if not data_lists:
            return
        for data_list in data_lists:
            # not match
            bench_name = api_data.get_data_by_header(CompareConst.BENCH_NAME, data_list)
            if not bench_name or bench_name == CompareConst.N_A:
                continue

            npu_shape = api_data.get_data_by_header(CompareConst.NPU_SHAPE, data_list)
            bench_shape = api_data.get_data_by_header(CompareConst.BENCH_SHAPE, data_list)

            if npu_shape is None or bench_shape is None:
                continue

            if npu_shape != bench_shape:
                api_data.set_result(data_list, self.result_level)
                api_data.set_err_msg(data_list, self.err_msg)


class RelativeWarnChecker(BaseAlgorithm):
    """
    适用于统计数据模式
    指标需要结合输入和输出共同计算得到
    一个 API 或模块的 output 相对误差是 input 相对误差的10倍，标记为 warning，默认选取norm观测
    """

    def __init__(self):
        self.threshold = 10
        self.result_level = ResultLevel.WARNING
        self.err_msg = (f'{self.result_level.value}: The norm relative error of output '
                        f'is {self.threshold} times that of input.')

    def run(self, api_data: ApiData, ignore_info: IgnoreInfo):
        if ignore_info in [IgnoreInfo.ALL_IGNORE, IgnoreInfo.INPUT_IGNORE]:
            return

        if not api_data.output_data:
            return

        norm_relative_error_max_in = abs(api_data.get_min_or_max_value(CompareConst.NORM_RELATIVE_ERR, is_min=False))
        norm_relative_error_max_out = abs(
            api_data.get_min_or_max_value(CompareConst.NORM_RELATIVE_ERR, is_input=False, is_min=False))

        should_set = False

        if norm_relative_error_max_in == 0:
            if norm_relative_error_max_out > 0.1:
                should_set = True
        elif norm_relative_error_max_out / norm_relative_error_max_in > self.threshold:
            should_set = True

        if should_set:
            api_data.set_result(api_data.output_data[0], self.result_level)
            api_data.set_err_msg(api_data.output_data[0], self.err_msg)


class CosineWarnChecker(BaseAlgorithm):
    """
    适用于真实数据模式
    指标需要结合输入和输出共同计算得到
    一个 API 或模块的 Cosine 的 input/parameters > 0.9 且 input/parameters - output > 0.1
    """

    def __init__(self):
        self.input_threshold = 0.9
        self.output_threshold = 0.1
        self.result_level = ResultLevel.WARNING
        self.err_msg = (f'{self.result_level.value}: The input/parameters of Cosine > {self.input_threshold}, '
                        f'and input/parameters - output > {self.output_threshold}')

    def run(self, api_data: ApiData, ignore_info: IgnoreInfo):
        if ignore_info in [IgnoreInfo.ALL_IGNORE, IgnoreInfo.INPUT_IGNORE]:
            return

        if not api_data.output_data:
            return

        min_input_cosine = api_data.get_min_or_max_value(CompareConst.COSINE)
        min_output_cosine = api_data.get_min_or_max_value(CompareConst.COSINE, is_input=False)

        if min_input_cosine > self.input_threshold and min_input_cosine - min_output_cosine > self.output_threshold:
            api_data.set_result(api_data.output_data[0], self.result_level)
            api_data.set_err_msg(api_data.output_data[0], self.err_msg)


TENSOR_CHECKERS = [InfNanErrChecker, OneThousandthErrChecker, RequiresGradErrChecker, ParametersErrChecker,
                   DTypeErrChecker, ShapeErrChecker, CosineWarnChecker]
STATISTICS_CHECKERS = [InfNanErrChecker, RelativeErrChecker, RequiresGradErrChecker, ParametersErrChecker,
                       DTypeErrChecker, ShapeErrChecker, RelativeWarnChecker]
MD5_CHECKERS = [CRC32ErrChecker]
STATISTICS_CHECKERS_PARALLEL_MERGE = [InfNanErrChecker, RelativeErrChecker, RequiresGradErrChecker,
                                      ParametersErrChecker, DTypeErrChecker, RelativeWarnChecker]



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import inspect
import os
import random
import sys
import types

import mindspore as ms
from mindspore import ops
from mindspore.common.jit_config import JitConfig
from mindspore.mint import nn

from msprobe.core.common.const import Const
from msprobe.core.common.decorator import recursion_depth_decorator
from msprobe.core.common.exceptions import DistributedNotInitializedError
from msprobe.core.common.file_utils import path_len_exceeds_limit, check_path_exists, save_npy
from msprobe.core.common.log import logger
from msprobe.core.common.utils import CompareException, check_seed_all, is_save_variable_valid
from msprobe.mindspore.common.const import Const as MsConst

try:
    from mindspore._c_expression import _set_init_iter
except ImportError:
    enable_dynamic_kbyk_dump = False
else:
    enable_dynamic_kbyk_dump = True

mindtorch_check_result = None
register_backward_hook_functions = {}
kwargs_exist_in_forward_hook = None
is_output_of_backward_hook_a_view = None


class MsprobeStep(ms.train.Callback):
    def __init__(self, debugger):
        super(MsprobeStep, self).__init__()
        self.debugger = debugger

    def on_train_begin(self, run_context):
        self.debugger.start()
        if enable_dynamic_kbyk_dump:
            _set_init_iter(0)

    def on_train_step_begin(self, run_context):
        self.debugger.start()

    def on_train_step_end(self, run_context):
        self.debugger.stop()
        self.debugger.step()


class MsprobeInitStep(ms.train.Callback):
    def on_train_begin(self, run_context):
        try:
            from ms._c_expression import _set_init_iter
        except ImportError:
            logger.warning('MsprobeInitStep does not work on this version of MindSpore.')
            return
        cb_params = run_context.original_args()
        _set_init_iter(cb_params.cur_step_num)


def get_rank_if_initialized():
    if ms.communication.GlobalComm.INITED:
        return ms.communication.get_rank()
    else:
        raise DistributedNotInitializedError("mindspore distributed environment is not initialized")


def convert_bf16_to_fp32(tensor):
    if tensor.dtype == ms.bfloat16:
        tensor = tensor.to(ms.float32)
    return tensor


def save_tensor_as_npy(tensor, file_path):
    if not path_len_exceeds_limit(file_path):
        tensor = convert_bf16_to_fp32(tensor)
        saved_tensor = tensor.asnumpy()
        save_npy(saved_tensor, file_path)
    else:
        logger.warning(f'The file path {file_path} length exceeds limit.')


def convert_to_int(value):
    if isinstance(value, bool):
        logger.error('The value in rank_id or step should be int, please check!')
        raise CompareException(CompareException.INVALID_OBJECT_TYPE_ERROR)
    try:
        return int(value)
    except Exception:
        return -1


def clean_input_kwargs(cell):
    if hasattr(cell, 'msprobe_input_kwargs'):
        del cell.msprobe_input_kwargs


def list_lowest_level_directories(root_dir):
    check_path_exists(root_dir)
    lowest_level_dirs = []

    def recurse_dirs(current_dir, depth=0):
        if depth > Const.MAX_DEPTH:
            logger.error(f'The directory {current_dir} has more than {Const.MAX_DEPTH} levels.')
            raise CompareException(CompareException.RECURSION_LIMIT_ERROR)
        for entry in os.listdir(current_dir):
            full_path = os.path.join(current_dir, entry)
            if os.path.isdir(full_path):
                if any(os.path.isdir(os.path.join(full_path, subentry)) for subentry in os.listdir(full_path)):
                    recurse_dirs(full_path, depth=depth + 1)
                else:
                    lowest_level_dirs.append(full_path)

    recurse_dirs(root_dir)
    return lowest_level_dirs


def seed_all(seed=1234, mode=False, rm_dropout=False):
    check_seed_all(seed, mode, rm_dropout)
    os.environ['PYTHONHASHSEED'] = str(seed)
    ms.set_seed(seed)
    random.seed(seed)
    ms.set_context(deterministic="ON" if mode else "OFF")
    os.environ['HCCL_DETERMINISTIC'] = str(mode)
    if rm_dropout:
        remove_dropout()


class Dropout(ops.Dropout):
    def __init__(self, keep_prob=0.5, seed0=0, seed1=1):
        super().__init__(1., seed0, seed1)


class Dropout2D(ops.Dropout2D):
    def __init__(self, keep_prob=0.5):
        super().__init__(1.)


class Dropout3D(ops.Dropout3D):
    def __init__(self, keep_prob=0.5):
        super().__init__(1.)


class DropoutExt(nn.Dropout):
    def __init__(self, p=0.5):
        super().__init__(0)


def dropout_ext(input_tensor, p=0.5, training=True):
    return input_tensor


def remove_dropout():
    ops.Dropout = Dropout
    ops.operations.Dropout = Dropout
    ops.Dropout2D = Dropout2D
    ops.operations.Dropout2D = Dropout2D
    ops.Dropout3D = Dropout3D
    ops.operations.Dropout3D = Dropout3D
    nn.Dropout = DropoutExt
    nn.functional.dropout = dropout_ext


def is_mindtorch():
    global mindtorch_check_result
    if mindtorch_check_result is None:
        mindtorch_check_result = False
        if 'torch' not in sys.modules:
            return mindtorch_check_result
        try:
            import torch
        except ImportError:
            return mindtorch_check_result
        tensor = torch.tensor(0.0)
        if isinstance(tensor, ms.Tensor):
            mindtorch_check_result = True
    return mindtorch_check_result


def set_register_backward_hook_functions():
    global register_backward_hook_functions
    if register_backward_hook_functions:
        return

    if is_mindtorch():
        import torch
        from msprobe.mindspore.dump.mindtorch import (
            _call_impl,
            register_full_backward_pre_hook,
            register_full_backward_hook
        )
        if not hasattr(torch.nn.Module, "register_full_backward_hook"):
            setattr(torch.nn.Module, "_call_impl", _call_impl)
            setattr(torch.nn.Module, "register_full_backward_pre_hook", register_full_backward_pre_hook)
            setattr(torch.nn.Module, "register_full_backward_hook", register_full_backward_hook)
        register_backward_hook_functions["pre"] = torch.nn.Module.register_full_backward_pre_hook
        register_backward_hook_functions["full"] = torch.nn.Module.register_full_backward_hook
    else:
        register_backward_hook_functions["pre"] = ms.nn.Cell.register_backward_pre_hook
        register_backward_hook_functions["full"] = ms.nn.Cell.register_backward_hook


def check_save_param(variable, name, save_backward):
    # try catch this api to skip invalid call
    valid_data_types = (ms.Tensor, int, float, str)
    if not is_save_variable_valid(variable, valid_data_types):
        valid_data_types_with_nested_types = valid_data_types + (dict, tuple, list)
        logger.warning("PrecisionDebugger.save variable type not valid, "
                       f"should be one of {valid_data_types_with_nested_types}"
                       "Skip current save process.")
        raise ValueError
    if not isinstance(name, str):
        logger.warning("PrecisionDebugger.save name not valid, "
                       "should be string. "
                       "skip current save process.")
        raise ValueError
    if not isinstance(save_backward, bool):
        logger.warning("PrecisionDebugger.save_backward name not valid, "
                       "should be bool. "
                       "Skip current save process.")
        raise ValueError


def is_graph_mode_cell_dump_allowed(config):
    if config.task not in [Const.TENSOR, Const.STATISTICS] or is_mindtorch() or not hasattr(ops, 'DumpGradient'):
        return False
    valid_mix_level = [MsConst.CELL_AND_API, Const.LEVEL_MIX]
    if config.level in valid_mix_level and config.execution_mode == MsConst.PYNATIVE_MODE:
        return True
    return config.level == MsConst.CELL or config.level == Const.LEVEL_L0


@recursion_depth_decorator('msprobe.mindspore.common.utils.is_decorated_by_jit')
def is_decorated_by_jit(func):
    closure = getattr(func, '__closure__', [])
    if closure:
        for obj in closure:
            if isinstance(obj.cell_contents, JitConfig):
                return True
            elif isinstance(obj.cell_contents, types.FunctionType) and hasattr(obj.cell_contents, '__closure__'):
                if is_decorated_by_jit(obj.cell_contents):
                    return True
    return False


@recursion_depth_decorator('msprobe.mindspore.common.utils.get_cells_and_names')
def get_cells_and_names(model, cells_set=None, name_prefix='', parent_cell=None):
    cells_set = cells_set if cells_set else set()
    if model in cells_set:
        return

    cells_set.add(model)
    jit_decorated = is_decorated_by_jit(model.construct)
    yield name_prefix, model, jit_decorated, parent_cell
    if jit_decorated:
        return

    children_cells = getattr(model, '_cells')
    for name, cell in children_cells.items():
        if cell:
            cells_name_prefix = f'{name_prefix}{Const.SEP}{name}' if name_prefix else name
            jit_decorated = is_decorated_by_jit(model.construct)
            if jit_decorated:
                yield cells_name_prefix, cell, jit_decorated, model
            else:
                for ele in get_cells_and_names(cell, cells_set, cells_name_prefix, model):
                    yield ele


def get_cells_and_names_with_index(models):
    cells_with_index_in_pynative_mode = {}
    cells_with_index_in_graph_mode = {}

    def distinguish_cells(cells):
        cells_in_pynative_mode = []
        cells_in_graph_mode = []
        for name, cell, jit_decorated, parent_cell in cells:
            if jit_decorated:
                cells_in_graph_mode.append((name, cell, parent_cell))
            else:
                cells_in_pynative_mode.append((name, cell))
        return cells_in_pynative_mode, cells_in_graph_mode

    if is_mindtorch():
        if isinstance(models, (list, tuple)):
            for index, model in enumerate(models):
                cells_with_index_in_pynative_mode[str(index)] = model.named_modules()
        else:
            cells_with_index_in_pynative_mode["-1"] = models.named_modules()
    else:
        if isinstance(models, (list, tuple)):
            for index, model in enumerate(models):
                cells = get_cells_and_names(model)
                cells_in_pynative_mode, cells_in_graph_mode = distinguish_cells(cells)
                cells_with_index_in_pynative_mode[str(index)] = cells_in_pynative_mode
                cells_with_index_in_graph_mode[str(index)] = cells_in_graph_mode
        else:
            cells = get_cells_and_names(models)
            cells_in_pynative_mode, cells_in_graph_mode = distinguish_cells(cells)
            cells_with_index_in_pynative_mode["-1"] = cells_in_pynative_mode
            cells_with_index_in_graph_mode["-1"] = cells_in_graph_mode

    return cells_with_index_in_pynative_mode, cells_with_index_in_graph_mode


def has_kwargs_in_forward_hook():
    global kwargs_exist_in_forward_hook

    if kwargs_exist_in_forward_hook is None:
        if is_mindtorch():
            kwargs_exist_in_forward_hook = True
            return kwargs_exist_in_forward_hook

        try:
            func_params = inspect.signature(nn.Cell.register_forward_hook).parameters
            kwargs_exist_in_forward_hook = 'with_kwargs' in func_params
        except Exception:
            kwargs_exist_in_forward_hook = False
        return kwargs_exist_in_forward_hook

    return kwargs_exist_in_forward_hook


def is_backward_hook_output_a_view():
    global is_output_of_backward_hook_a_view

    if is_output_of_backward_hook_a_view is None:
        is_output_of_backward_hook_a_view = False
        if getattr(ms, '__version__', '2.4.0') < '2.7.0':
            return is_output_of_backward_hook_a_view
        try:
            from mindspore.ops.operations import _inner_ops as inner
            call_func = getattr(inner.CellBackwardHook, '__call__')
            func_params = inspect.signature(call_func).parameters
        except Exception:
            return is_output_of_backward_hook_a_view
        if 'args' in func_params and func_params['args'].kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            is_output_of_backward_hook_a_view = True

    return is_output_of_backward_hook_a_view


def wrap_backward_hook_call_func(call_func):
    if not is_backward_hook_output_a_view():
        return call_func

    from mindspore.common.api import _pynative_executor as executor
    from mindspore._c_expression import CreationType

    def new_call(self, args):
        outputs = call_func(self, args)
        if isinstance(outputs, ms.Tensor):
            executor.set_creation_type(outputs, CreationType.DEFAULT)
        elif isinstance(outputs, tuple):
            for item in outputs:
                if isinstance(item, ms.Tensor):
                    executor.set_creation_type(item, CreationType.DEFAULT)
        return outputs

    new_call.__name__ = '__call__'

    return new_call


def cast_to_float_if_fp8(tensor):
    dtype = str(tensor.dtype)
    if dtype in MsConst.FP8_TYPE_LIST:
        logger.debug(
            f"The {dtype} tensor analyzing/saving is unsupported in dump function."
            f"Casting to a float for processing"
        )
        tensor = tensor.to(ms.float32)
    return tensor



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

__all__ = [
    "Rule",
    "NumberChecker",
    "PathChecker",
    "StringChecker",
    "ArgsChecker",
    "DictChecker",
    "ObjectChecker",
    "ListChecker",
    "validate_params",
]

from msprobe.infer.utils.check.number_checker import NumberChecker
from msprobe.infer.utils.check.path_checker import PathChecker
from msprobe.infer.utils.check.string_checker import StringChecker
from msprobe.infer.utils.check.args_checker import ArgsChecker
from msprobe.infer.utils.check.dict_checker import DictChecker
from msprobe.infer.utils.check.obj_checker import ObjectChecker
from msprobe.infer.utils.check.list_checker import ListChecker
from msprobe.infer.utils.check.func_wrapper import validate_params
from msprobe.infer.utils.check.rule import Rule
